"""
Implementation of classes containing the estimation results.

Michel Bierlaire
Sun Sep 29 16:54:42 2024
"""

import logging
from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from typing import Any

import numpy as np
from yaml import (
    Dumper,
    Loader,
    Node,
    SafeLoader,
    add_constructor,
    add_representer,
    dump,
    load,
)

from biogeme.tools import safe_deserialize_array
from biogeme.version import get_version, versionDate

logger = logging.getLogger(__name__)


@dataclass
class RawEstimationResults:
    """Data class containing the unprocessed estimation results. Must be simple and contain no method, in order to
    serialize it easily."""

    model_name: str
    user_notes: str
    beta_names: list[str]
    beta_values: list[float]
    lower_bounds: list[float]
    upper_bounds: list[float]
    gradient: list[float]
    hessian: list[list[float]]
    bhhh: list[list[float]]
    null_log_likelihood: float
    initial_log_likelihood: float | None
    final_log_likelihood: float
    data_name: str
    sample_size: int
    number_of_observations: int
    monte_carlo: bool
    number_of_draws: int
    types_of_draws: dict[str, str]
    number_of_excluded_data: int
    draws_processing_time: timedelta
    optimization_messages: dict[str, Any]
    convergence: bool
    bootstrap: list[list[float]]
    bootstrap_time: timedelta | None


# Custom handlers to serialize and deserialize complex objects
def timedelta_representer(dumper: Dumper, data: timedelta) -> Node:
    """Represent a timedelta object as the total seconds in a YAML string."""
    return dumper.represent_str(str(data.total_seconds()))


def timedelta_constructor(loader: Loader, node: Node) -> timedelta:
    """Construct a timedelta object from a YAML scalar representing total seconds."""
    value: str = loader.construct_scalar(node)
    return timedelta(seconds=float(value))


# Register the custom handlers with PyYAML
add_representer(timedelta, timedelta_representer)
add_constructor('tag:yaml.org,2002:str', timedelta_constructor)


def check_for_invalid_yaml_values(data: Any, path='root'):
    """Recursively checks for NaN or binary values in the data."""
    if isinstance(data, float) and np.isnan(data):
        raise ValueError(f'Invalid NaN value found at {path}')
    if isinstance(data, bytes):
        raise ValueError(f'Binary data (bytes) not allowed at {path}')
    if isinstance(data, dict):
        for key, value in data.items():
            check_for_invalid_yaml_values(value, f'{path}.{key}')
    elif isinstance(data, (list, tuple)):
        for i, item in enumerate(data):
            check_for_invalid_yaml_values(item, f'{path}[{i}]')


def contains_python_tags(yaml_string):
    return '!!python/' in yaml_string or '!!binary' in yaml_string


# To serialize the RawEstimationResults instance to a YAML file
def serialize_to_yaml(data: RawEstimationResults, filename: str) -> None:
    """Dump the data in an ASCII file

    :param data: raw estimation results
    :param filename: name of the file
    """
    dict_data = asdict(data)
    check_for_invalid_yaml_values(dict_data)
    # Convert to YAML string
    yaml_string = dump(dict_data)

    # Check for unsafe Python-specific tags
    if contains_python_tags(yaml_string):
        raise ValueError(
            f"The YAML output [{yaml_string}] contains unsafe Python object tags. Aborting serialization of {filename}."
        )

    with open(filename, 'w') as file:
        now = datetime.now()
        print(f'# File {filename} has automatically been generated on {now}', file=file)
        print(f'# biogeme {get_version()} [{versionDate}]\n', file=file)
        file.write(yaml_string)
    logger.info(f'File {filename} has been generated.')


# To deserialize the RawEstimationResults instance from a YAML file
def deserialize_from_yaml(filename) -> RawEstimationResults:
    """Restore data from a YAML file

    :param filename: name of the file
    :return: raw estimation results
    """
    with open(filename, 'r') as file:
        data = load(file, Loader=SafeLoader)
    bootstrap_time = (
        timedelta(seconds=float(data['bootstrap_time']))
        if data['bootstrap_time'] is not None
        else None
    )
    return RawEstimationResults(
        model_name=data['model_name'],
        user_notes=data['user_notes'],
        beta_names=list(data['beta_names']),
        beta_values=list(data['beta_values']),
        lower_bounds=data['lower_bounds'],
        upper_bounds=data['upper_bounds'],
        gradient=safe_deserialize_array(data['gradient']),
        hessian=safe_deserialize_array(data['hessian']),
        bhhh=safe_deserialize_array(data['bhhh']),
        null_log_likelihood=data['null_log_likelihood'],
        initial_log_likelihood=data['initial_log_likelihood'],
        final_log_likelihood=data['final_log_likelihood'],
        data_name=data['data_name'],
        sample_size=data['sample_size'],
        number_of_observations=data['number_of_observations'],
        monte_carlo=data['monte_carlo'],
        number_of_draws=data['number_of_draws'],
        types_of_draws=data['types_of_draws'],
        number_of_excluded_data=data['number_of_excluded_data'],
        draws_processing_time=timedelta(seconds=float(data['draws_processing_time'])),
        optimization_messages=data['optimization_messages'],
        convergence=data['convergence'],
        bootstrap=data['bootstrap'],
        bootstrap_time=bootstrap_time,
    )



"""Audit the dataframe"""

from typing import NamedTuple

import numpy as np
import pandas as pd

from biogeme.audit_tuple import AuditTuple
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Expression
from biogeme.tools import count_number_of_groups
from .container import Database


class ChosenAvailable(NamedTuple):
    chosen: int
    available: int


def check_availability_of_chosen_alt(
    database: Database, avail: dict[int:Expression], choice: Expression
) -> pd.Series:
    """Check if the chosen alternative is available for each entry
    in the database.

    :param database: object containing the data
    :param avail: list of expressions to evaluate the
                  availability conditions for each alternative.
    :param choice: expression for the chosen alternative.

    :return: numpy series of bool, long as the number of entries
             in the database, containing True is the chosen alternative is
             available, False otherwise.

    :raise BiogemeError: if the chosen alternative does not appear
        in the availability dict
    """
    from biogeme.calculator import evaluate_expression

    choice_array = evaluate_expression(
        expression=choice,
        numerically_safe=False,
        database=database,
    )
    calculated_avail = {}
    for key, expression in avail.items():
        calculated_avail[key] = evaluate_expression(
            expression=expression,
            numerically_safe=False,
            database=database,
        )
    try:
        avail_chosen = np.array(
            [calculated_avail[c][i] for i, c in enumerate(choice_array)]
        )
        return avail_chosen != 0
    except KeyError as exc:
        for c in choice_array:
            if c not in calculated_avail:
                err_msg = (
                    f'Chosen alternative {c} does not appear in '
                    f'availability dict: {calculated_avail.keys()}'
                )
                raise BiogemeError(err_msg) from exc


def choice_availability_statistics(
    database: Database, avail: dict[int:Expression], choice: Expression
) -> dict[int, ChosenAvailable]:
    """Calculates the number of times an alternative is chosen and available

    :param database: object containing the data
    :param avail: list of expressions to evaluate the
                  availability conditions for each alternative.
    :param choice: expression for the chosen alternative.

    :return: for each alternative, a tuple containing the number of time
        it is chosen, and the number of time it is available.

    :raise BiogemeError: if the database is empty.
    """
    from biogeme.calculator import evaluate_expression

    choice_array = evaluate_expression(
        expression=choice,
        numerically_safe=False,
        database=database,
    )
    calculated_avail = {}
    for key, expression in avail.items():
        calculated_avail[key] = evaluate_expression(
            expression=expression,
            numerically_safe=False,
            database=database,
        )
    unique = np.unique(choice_array, return_counts=True)
    choice_stat = {alt: int(unique[1][i]) for i, alt in enumerate(list(unique[0]))}
    avail_stat = {k: sum(a) for k, a in calculated_avail.items()}
    the_results = {
        alt: ChosenAvailable(chosen=c, available=avail_stat[alt])
        for alt, c in choice_stat.items()
    }
    return the_results


def audit_dataframe(data: pd.DataFrame) -> AuditTuple:
    """
    Performs a series of checks and reports warnings and errors for a pandas DataFrame.

    :param data: The DataFrame to audit.
    :return: the list of errors.
    """
    list_of_warnings = []
    list_of_errors = []

    for col, dtype in data.dtypes.items():
        if not np.issubdtype(dtype, np.number):
            list_of_errors.append(
                f'Column {col} in the database contains non-numeric type: {dtype}'
            )

    if data.isnull().values.any():
        nan_locations = data.isnull()
        rows_with_nan = data.index[nan_locations.any(axis=1)].tolist()
        cols_with_nan = data.columns[nan_locations.any(axis=0)].tolist()
        list_of_errors.append(
            f"The database contains NaN value(s).\n"
            f"Columns with NaN: {cols_with_nan}\n"
            f"Rows with NaN: {rows_with_nan}\n"
            f"Use database.dataframe.isna() to inspect further."
        )

    return AuditTuple(errors=list_of_errors, warnings=list_of_warnings)


def audit_panel_dataframe(
    data: pd.DataFrame, id_column: str
) -> tuple[list[str], list[str]]:
    """
    Performs panel-specific checks on a pandas DataFrame, ensuring entries for
    the same individual are contiguous.

    :param data: The DataFrame to audit.
    :param id_column: The name of the column identifying individuals.
    :return: A tuple (list_of_errors, list_of_warnings).
    """
    list_of_errors = []
    list_of_warnings = []

    if id_column not in data.columns:
        list_of_errors.append(
            f"The column '{id_column:d}' is missing from the dataset."
        )
        return list_of_errors, list_of_warnings

    original_groups = count_number_of_groups(data, id_column)
    sorted_data = data.sort_values(by=id_column).reset_index(drop=True)
    sorted_groups = count_number_of_groups(sorted_data, id_column)

    if original_groups != sorted_groups:
        list_of_errors.append(
            f"The data must be sorted so that entries for the same individual "
            f"are contiguous. Found {original_groups} original groups, "
            f"but {sorted_groups} after sorting."
        )

    return list_of_errors, list_of_warnings



"""Arithmetic expressions accepted by Biogeme: generic class

Michel Bierlaire
Tue Mar 25 17:00:02 2025
"""

from __future__ import annotations

import logging
from itertools import chain
from typing import NamedTuple, TYPE_CHECKING, TypeAlias

from biogeme.exceptions import BiogemeError, NotImplementedError

from .elementary_types import TypeOfElementaryExpression
from .jax_utils import JaxFunctionType
from .numeric_tools import is_numeric
from .validation import validate_expression_type

if TYPE_CHECKING:
    from .elementary_expressions import Elementary

logger = logging.getLogger(__name__)


class LogitTuple(NamedTuple):
    choice: Expression
    availabilities: dict[int, Expression]


class Expression:
    """This is the general arithmetic expression in biogeme.
    It serves as a base class for concrete expressions.
    """

    def __init__(self) -> None:
        """Constructor"""
        self.children = []  #: List of children expressions

        self.fixed_beta_values = None
        """values of the Beta that are not estimated
        """

    def __bool__(self) -> None:
        error_msg = f'Expression {str(self)} cannot be used in a boolean expression. Use & for "and" and | for "or"'
        raise BiogemeError(error_msg)

    def get_value(self) -> float:
        """Calculates the value of the expression if it is simple"""
        raise BiogemeError(
            f'Expression of type {self.get_class_name()} does not support direct value evaluation.'
        )

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates recursively a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """
        raise NotImplementedError(f'{repr(self)}')

    def deep_flat_copy(self) -> Expression:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        """
        raise NotImplementedError(f'{repr(self)}')

    def __repr__(self) -> str:
        """built-in function used to compute the 'official' string reputation
        of an object

        :return: description of the expression

        """
        return self.__str__()

    def __add__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for addition.

        :param other: expression to be added

        :return: self + other

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .plus import Plus

        return Plus(self, other)

    def __radd__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for addition.

        :param other: expression to be added
        :param other: expression to be added
        :type other: biogeme.expressions.Expression

        :return: other + self
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.

        """
        validate_expression_type(other)
        from .plus import Plus

        return Plus(other, self)

    def __sub__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for subtraction.

        :param other: expression to subtract
        :type other: biogeme.expressions.Expression

        :return: self - other
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.

        """
        validate_expression_type(other)
        from .minus import Minus

        return Minus(self, other)

    def __rsub__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for subtraction.

        :param other: expression to be subtracted
        :type other: biogeme.expressions.Expression

        :return: other - self
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.

        """
        validate_expression_type(other)
        from .minus import Minus

        return Minus(other, self)

    def __mul__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for multiplication.

        :param other: expression to be multiplied
        :type other: biogeme.expressions.Expression

        :return: self * other
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.

        """
        validate_expression_type(other)
        from .times import Times

        return Times(self, other)

    def __rmul__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for multiplication.

        :param other: expression to be multiplied
        :type other: biogeme.expressions.Expression

        :return: other * self
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .times import Times

        return Times(other, self)

    def __truediv__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for division.

        :param other: expression for division
        :type other: biogeme.expressions.Expression

        :return: self / other
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .divide import Divide

        return Divide(self, other)

    def __rtruediv__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for division.

        :param other: expression for division
        :type other: biogeme.expressions.Expression

        :return: other / self
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .divide import Divide

        return Divide(other, self)

    def __neg__(self) -> Expression:
        """
        Operator overloading. Generate an expression for unary minus.

        :return: -self
        :rtype: biogeme.expressions.Expression
        """
        from .unary_minus import UnaryMinus

        return UnaryMinus(self)

    def __pow__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for power.

        :param other: expression for power
        :type other: biogeme.expressions.Expression

        :return: self ^ other
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.

        """
        validate_expression_type(other)
        if is_numeric(other):
            from .power_constant import PowerConstant

            return PowerConstant(child=self, exponent=float(other))

        from .numeric_expressions import Numeric

        if isinstance(other, Numeric):
            from .power_constant import PowerConstant

            return PowerConstant(self, other.get_value())

        if isinstance(other, Expression):
            from .power import Power

            return Power(self, other)

    def __rpow__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for power.

        :param other: expression for power
        :type other: biogeme.expressions.Expression

        :return: other ^ self
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .power import Power

        return Power(other, self)

    def __and__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for logical and.

        :param other: expression for logical and
        :type other: biogeme.expressions.Expression

        :return: self and other
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .logical_and import And

        return And(self, other)

    def __rand__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for logical and.

        :param other: expression for logical and
        :type other: biogeme.expressions.Expression

        :return: other and self
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .logical_and import And

        return And(other, self)

    def __or__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for logical or.

        :param other: expression for logical or
        :type other: biogeme.expressions.Expression

        :return: self or other
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .logical_or import Or

        return Or(self, other)

    def __ror__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for logical or.

        :param other: expression for logical or
        :type other: biogeme.expressions.Expression

        :return: other or self
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .logical_or import Or

        return Or(other, self)

    def __eq__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for comparison.

        :param other: expression for equality
        :type other: biogeme.expressions.Expression

        :return: self == other
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .comparison_expressions import Equal

        return Equal(self, other)

    def __ne__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for comparison.

        :param other: expression for difference
        :type other: biogeme.expressions.Expression

        :return: self != other
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .comparison_expressions import NotEqual

        return NotEqual(self, other)

    def __le__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for comparison.

        :param other: expression for less or equal
        :type other: biogeme.expressions.Expression

        :return: self <= other
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .comparison_expressions import LessOrEqual

        return LessOrEqual(self, other)

    def __ge__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for comparison.

        :param other: expression for greater or equal
        :type other: biogeme.expressions.Expression

        :return: self >= other
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .comparison_expressions import GreaterOrEqual

        return GreaterOrEqual(self, other)

    def __lt__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for comparison.

        :param other: expression for less than
        :type other: biogeme.expressions.Expression

        :return: self < other
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .comparison_expressions import Less

        return Less(self, other)

    def __gt__(self, other: ExpressionOrNumeric) -> Expression:
        """
        Operator overloading. Generate an expression for comparison.

        :param other: expression for greater than
        :type other: biogeme.expressions.Expression

        :return: self > other
        :rtype: biogeme.expressions.Expression

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        """
        validate_expression_type(other)
        from .comparison_expressions import Greater

        return Greater(self, other)

    def dict_of_draw_types(self) -> dict[str:str]:
        """Extract a dict containing the types of draws involved in the expression"""
        the_draws = self.dict_of_elementary_expression(
            the_type=TypeOfElementaryExpression.DRAWS
        )
        return {name: expression.drawType for name, expression in the_draws.items()}

    def logit_choice_avail(self) -> list[LogitTuple]:
        """Extract a dict with all elementary expressions of a specific type

        :return: returns a dict with the variables appearing in the
               expression the keys being their names.
        :rtype: dict(string:biogeme.expressions.Expression)

        """
        result: list[LogitTuple] = list(
            chain.from_iterable(e.logit_choice_avail() for e in self.children)
        )
        return result

    def get_elementary_expression(self, name: str) -> Elementary | None:
        """Return: an elementary expression from its name if it appears in the
        expression.

        :param name: name of the elementary expression.
        :type name: string

        :return: the expression if it exists. None otherwise.
        :rtype: biogeme.expressions.Expression
        """
        for e in self.get_children():
            if e.get_elementary_expression(name) is not None:
                return e.get_elementary_expression(name)
        return None

    def get_class_name(self) -> str:
        """
        Obtain the name of the top class of the expression structure

        :return: the name of the class
        :rtype: string
        """
        n = type(self).__name__
        return n

    def get_children(self) -> list[Expression]:
        """Retrieve the list of children

        :return: list of children
        :rtype: list(Expression)
        """
        return self.children

    def embed_expression(self, name: type) -> bool:
        """Check if an expression embeds a specific operator"""
        if isinstance(self, name):
            return True
        return any(child.embed_expression(name) for child in self.get_children())

    def set_specific_id(
        self, the_name, specific_id, the_type: TypeOfElementaryExpression
    ):
        """The elementary IDs identify the position of each element in the corresponding datab"""
        for child in self.get_children():
            child.set_specific_id(the_name, specific_id, the_type)

    def set_maximum_number_of_observations_per_individual(
        self, max_number: int
    ) -> None:
        for child in self.get_children():
            child.set_maximum_number_of_observations_per_individual(
                max_number=max_number
            )

    def change_init_values(self, betas: dict[str, float]):
        """Modifies the initial values of the Beta parameters.

        The fact that the parameters are fixed or free is irrelevant here.

        :param betas: dictionary where the keys are the names of the
                      parameters, and the values are the new value for
                      the parameters.
        :type betas: dict(string:float)
        """
        for child in self.get_children():
            child.change_init_values(betas=betas)

    def requires_draws(self):
        from .draws import Draws

        return self.embed_expression(Draws)


ExpressionOrNumeric: TypeAlias = Expression | float | int | bool



import logging
from dataclasses import dataclass

import pandas as pd

from biogeme.calculator import CompiledFormulaEvaluator, MultiRowEvaluator
from biogeme.default_parameters import ParameterValue
from biogeme.likelihood import AlgorithmResults, model_estimation
from biogeme.model_elements import ModelElements
from biogeme.optimization import OptimizationAlgorithm
from .split_databases import EstimationValidationModels, split_databases

logger = logging.getLogger(__name__)


@dataclass
class ValidationResult:
    estimation_modeling_elements: ModelElements
    validation_modeling_elements: ModelElements
    simulated_values: pd.DataFrame


def cross_validate_model(
    the_algorithm: OptimizationAlgorithm,
    modeling_elements: ModelElements,
    parameters: dict[str, ParameterValue],
    starting_values: dict[str, float],
    slices: int,
    numerically_safe: bool,
    groups: str | None = None,
) -> list[ValidationResult]:
    validation_models: list[EstimationValidationModels] = split_databases(
        model_elements=modeling_elements, slices=slices, groups=groups
    )
    results = []
    for i, fold in enumerate(validation_models, 1):
        # Estimation phase
        the_function_evaluator = CompiledFormulaEvaluator(
            model_elements=fold.estimation,
            second_derivatives_mode=parameters['calculating_second_derivatives'],
            numerically_safe=numerically_safe,
        )
        one_result: AlgorithmResults = model_estimation(
            the_algorithm=the_algorithm,
            function_evaluator=the_function_evaluator,
            parameters=parameters,
            some_starting_values=starting_values,
            save_iterations_filename=None,
        )
        estimated_betas = fold.estimation.expressions_registry.get_named_betas_values(
            values=one_result.solution
        )
        simulation_evaluator = MultiRowEvaluator(
            model_elements=fold.validation, numerically_safe=numerically_safe
        )
        simulated_values: pd.DataFrame = simulation_evaluator.evaluate(
            the_betas=estimated_betas
        )
        result = ValidationResult(
            estimation_modeling_elements=fold.estimation,
            validation_modeling_elements=fold.validation,
            simulated_values=simulated_values,
        )
        results.append(result)

    return results



"""Function to check and process numeric expressions

:author: Michel Bierlaire
:date: Sat Sep  9 15:27:17 2023
"""

import logging

logger = logging.getLogger(__name__)


def is_numeric(obj) -> bool:
    """Checks if an object is numeric
    :param obj: obj to be checked
    :type obj: Object

    """
    return isinstance(obj, (int, float, bool))



"""

Model specification
===================

Script containing the model specification.

:author: Michel Bierlaire
:date: Wed Nov  1 17:37:33 2023
"""
from biogeme.expressions import Beta, Variable, log
from biogeme.sampling_of_alternatives import CrossVariableTuple

# %%
# Variable combining attributes of the alternatives and
# characteristics of the decision-maker
combined_variables = [
    CrossVariableTuple(
        'log_dist',
        log(
            (
                (Variable('user_lat') - Variable('rest_lat')) ** 2
                + (Variable('user_lon') - Variable('rest_lon')) ** 2
            )
            ** 0.5
        ),
    )
]

# %%
# Parameters to estimate.
beta_rating = Beta('beta_rating', 0, None, None, 0)
beta_price = Beta('beta_price', 0, None, None, 0)
beta_chinese = Beta('beta_chinese', 0, None, None, 0)
beta_japanese = Beta('beta_japanese', 0, None, None, 0)
beta_korean = Beta('beta_korean', 0, None, None, 0)
beta_indian = Beta('beta_indian', 0, None, None, 0)
beta_french = Beta('beta_french', 0, None, None, 0)
beta_mexican = Beta('beta_mexican', 0, None, None, 0)
beta_lebanese = Beta('beta_lebanese', 0, None, None, 0)
beta_ethiopian = Beta('beta_ethiopian', 0, None, None, 0)
beta_log_dist = Beta('beta_log_dist', 0, None, None, 0)

# %%
# Utility function.
V = (
    beta_rating * Variable('rating')
    + beta_price * Variable('price')
    + beta_chinese * Variable('category_Chinese')
    + beta_japanese * Variable('category_Japanese')
    + beta_korean * Variable('category_Korean')
    + beta_indian * Variable('category_Indian')
    + beta_french * Variable('category_French')
    + beta_mexican * Variable('category_Mexican')
    + beta_lebanese * Variable('category_Lebanese')
    + beta_ethiopian * Variable('category_Ethiopian')
    + beta_log_dist * Variable('log_dist')
)



"""Arithmetic expressions accepted by Biogeme: multiple sum

Michel Bierlaire
Sat Sep  9 15:29:36 2023
"""

from __future__ import annotations

import logging
from typing import Any, TYPE_CHECKING

import jax.numpy as jnp

from biogeme.exceptions import BiogemeError
from .base_expressions import Expression, ExpressionOrNumeric
from .convert import validate_and_convert
from .jax_utils import JaxFunctionType

if TYPE_CHECKING:
    pass

logger = logging.getLogger(__name__)


class MultipleSum(Expression):
    """This expression returns the sum of several other expressions.

    It is a generalization of 'Plus' for more than two terms
    """

    def __init__(
        self,
        list_of_expressions: list[ExpressionOrNumeric] | dict[Any:ExpressionOrNumeric],
    ):
        """Constructor

        :param list_of_expressions: list of objects representing the
                                     terms of the sum.

        :type list_of_expressions: list(biogeme.expressions.Expression)

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        :raise BiogemeError: if the list of expressions is empty
        :raise BiogemeError: if the list of expressions is neither a dict nor a list
        """
        if not list_of_expressions:
            raise BiogemeError('The argument of MultipleSum cannot be empty')

        super().__init__()

        if isinstance(list_of_expressions, dict):
            items = list_of_expressions.values()
        elif isinstance(list_of_expressions, list):
            items = list_of_expressions
        else:
            raise BiogemeError('Argument of MultipleSum must be a dict or a list.')

        for expression in items:
            self.children.append(validate_and_convert(expression))

    def deep_flat_copy(self) -> MultipleSum:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        copy_list_of_expressions = [
            expression.deep_flat_copy() for expression in self.children
        ]
        return type(self)(list_of_expressions=copy_list_of_expressions)

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        result = 0.0
        for e in self.get_children():
            result += e.get_value()
        return result

    def __str__(self) -> str:
        s = 'MultipleSum(' + ', '.join([f'{e}' for e in self.get_children()]) + ')'
        return s

    def __repr__(self) -> str:
        return f"MultipleSum({repr(self.get_children())})"

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        compiled_children = [
            child.recursive_construct_jax_function(numerically_safe=numerically_safe)
            for child in self.get_children()
        ]

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.ndarray:
            terms = [
                fn(parameters, one_row, the_draws, the_random_variables)
                for fn in compiled_children
            ]
            return jnp.sum(jnp.stack(terms))

        return the_jax_function



""" Implements various models.

:author: Michel Bierlaire
:date: Fri Mar 29 17:13:14 2019
"""

from __future__ import annotations

import logging

import biogeme.exceptions as excep
from biogeme.deprecated import deprecated
from biogeme.expressions import (
    Expression,
    ExpressionOrNumeric,
    exp,
    log,
    MultipleSum,
    Numeric,
    ConditionalTermTuple,
    ConditionalSum,
)
from biogeme.models import logmev, mev
from biogeme.nests import NestsForNestedLogit, OldNestsForNestedLogit

logger = logging.getLogger(__name__)


def get_mev_generating_for_nested(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForNestedLogit | OldNestsForNestedLogit,
) -> Expression:
    """Implements the  MEV generating function for the nested logit model

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :param nests: an object describing the nests

    :return: a dictionary mapping each alternative id with the function

    .. math:: G(e^{V_1},
              \\ldots,e^{V_J}) =  \\sum_m \\left( \\sum_{\\ell \\in C_m}
              y_\\ell^{\\mu_m}\\right)^{\\frac{\\mu}{\\mu_m}}

    where :math:`G` is the MEV generating function.

    """
    if not isinstance(nests, NestsForNestedLogit):
        logger.warning(
            'It is recommended to define the nests of the nested logit model using '
            'the objects OneNestForNestedLogit and NestsForNestedLogit defined '
            'in biogeme.nests.'
        )
        nests = NestsForNestedLogit(choice_set=list(util), tuple_of_nests=nests)
    ok, message = nests.check_partition()
    if not ok:
        raise excep.BiogemeError(message)

    terms_for_nests = []
    for m in nests:
        if availability is None:
            sum_terms = [exp(m.nest_param * util[i]) for i in m.list_of_alternatives]
            the_sum = MultipleSum(sum_terms)
        else:
            sum_terms = [
                ConditionalTermTuple(
                    condition=availability[i] != Numeric(0),
                    term=exp(m.nest_param * util[i]),
                )
                for i in m.list_of_alternatives
            ]
            the_sum = ConditionalSum(list_of_terms=sum_terms)
        terms_for_nests.append(the_sum ** (1.0 / m.nest_param))
    if nests.alone is not None:
        for i in nests.alone:
            terms_for_nests.append(util[i])
    return MultipleSum(terms_for_nests)


@deprecated(get_mev_generating_for_nested)
def getMevGeneratingForNested(
    util: dict[int, Expression],
    availability: dict[int, Expression],
    nests: NestsForNestedLogit | OldNestsForNestedLogit,
) -> Expression:
    pass


def get_mev_for_nested(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForNestedLogit | OldNestsForNestedLogit,
) -> dict[int, Expression]:
    """Implements the derivatives of MEV generating function for the
    nested logit model

    :param util: dict of objects representing the utility functions of
        each alternative, indexed by numerical ids.

    :param availability: dict of objects representing the availability of each
               alternative, indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :param nests: object containing the description of the nests.

    :return: a dictionary mapping each alternative id with the function

        .. math:: \\ln \\frac{\\partial G}{\\partial y_i}(e^{V_1},
              \\ldots,e^{V_J}) = e^{(\\mu_m-1)V_i}
              \\left(\\sum_{i=1}^{J_m} e^{\\mu_m V_i}\\right)^
              {\\frac{1}{\\mu_m}-1}

        where :math:`m` is the (only) nest containing alternative :math:`i`,
        and :math:`G` is the MEV generating function.

    """
    if not isinstance(nests, NestsForNestedLogit):
        logger.warning(
            'It is recommended to define the nests of the nested logit model using '
            'the objects OneNestForNestedLogit and NestsForNestedLogit defined '
            'in biogeme.nests.'
        )
        nests = NestsForNestedLogit(choice_set=list(util), tuple_of_nests=nests)

    ok, message = nests.check_partition()
    if not ok:
        raise excep.BiogemeError(message)

    if nests.alone is None:
        log_gi = {}
    else:
        log_gi = {i: Numeric(0) for i in nests.alone}
    for m in nests:
        if availability is None:
            sum_terms = [exp(m.nest_param * util[i]) for i in m.list_of_alternatives]
            the_sum = MultipleSum(sum_terms)
        else:
            sum_terms = [
                ConditionalTermTuple(
                    condition=availability[i] != Numeric(0),
                    term=exp(m.nest_param * util[i]),
                )
                for i in m.list_of_alternatives
            ]
            the_sum = ConditionalSum(list_of_terms=sum_terms)

        for i in m.list_of_alternatives:
            log_gi[i] = (m.nest_param - 1.0) * util[i] + (
                1.0 / m.nest_param - 1.0
            ) * log(the_sum)
    return log_gi


@deprecated(get_mev_for_nested)
def getMevForNested(
    V: dict[int, Expression],
    availability: dict[int, Expression] | None,
    nests: NestsForNestedLogit | OldNestsForNestedLogit,
) -> dict[int, Expression]:
    pass


def get_mev_for_nested_mu(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForNestedLogit | OldNestsForNestedLogit,
    mu: ExpressionOrNumeric,
) -> dict[int, Expression]:
    """Implements the MEV generating function for the nested logit model,
    including the scale parameter

    :param util: dict of objects representing the utility functions of
        each alternative, indexed by numerical ids.

    :param availability: dict of objects representing the availability
        of each alternative, indexed
        by numerical ids. Must be consistent with util, or
        None. In this case, all alternatives are supposed to be
        always available.

    :param nests: object describing the nesting structure

    :param mu: scale parameter

    :return: a dictionary mapping each alternative id with the function

        .. math:: \\frac{\\partial G}{\\partial y_i}(e^{V_1},\\ldots,e^{V_J}) =
                  \\mu e^{(\\mu_m-1)V_i} \\left(\\sum_{i=1}^{J_m}
                  e^{\\mu_m V_i}\\right)^{\\frac{\\mu}{\\mu_m}-1}

        where :math:`m` is the (only) nest containing alternative :math:`i`,
        and :math:`G` is the MEV generating function.

    """
    if not isinstance(nests, NestsForNestedLogit):
        logger.warning(
            'It is recommended to define the nests of the nested logit model using '
            'the objects OneNestForNestedLogit and NestsForNestedLogit defined '
            'in biogeme.nests.'
        )
        nests = NestsForNestedLogit(choice_set=list(util), tuple_of_nests=nests)
    ok, message = nests.check_partition()
    if not ok:
        raise excep.BiogemeError(message)

    if nests.alone is None:
        log_gi = {}
    else:
        log_gi = {i: log(mu) + (mu - 1) * util[i] for i in nests.alone}
    for m in nests:
        if availability is None:
            sum_terms = [exp(m.nest_param * util[i]) for i in m.list_of_alternatives]
            the_sum = MultipleSum(sum_terms)

        else:
            sum_terms = [
                ConditionalTermTuple(
                    condition=availability[i] != Numeric(0),
                    term=exp(m.nest_param * util[i]),
                )
                for i in m.list_of_alternatives
            ]
            the_sum = ConditionalSum(list_of_terms=sum_terms)
        for i in m.list_of_alternatives:
            log_gi[i] = (
                log(mu)
                + (m.nest_param - 1.0) * util[i]
                + (mu / m.nest_param - 1.0) * log(the_sum)
            )
    return log_gi


@deprecated(get_mev_for_nested_mu)
def getMevForNestedMu(
    util: dict[int, Expression],
    availability: dict[int, Expression] | None,
    nests: NestsForNestedLogit | OldNestsForNestedLogit,
    mu: Expression,
) -> dict[int, Expression]:
    pass


def nested(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForNestedLogit | OldNestsForNestedLogit,
    choice: ExpressionOrNumeric,
) -> Expression:
    """Implements the nested logit model as a MEV model.

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :param availability: dict of objects representing the availability
                         of each alternative, indexed by numerical
                         ids. Must be consistent with util, or None. In
                         this case, all alternatives are supposed to
                         be always available.

    :param nests: object containing the description of the nests.

    :param choice: id of the alternative for which the probability must be
              calculated.

    :return: choice probability for the nested logit model,
             based on the derivatives of the MEV generating function produced
             by the function getMevForNested

    :raise BiogemeError: if the definition of the nests is invalid.
    """

    log_gi = get_mev_for_nested(util, availability, nests)
    P = mev(util, log_gi, availability, choice)
    return P


def lognested(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForNestedLogit | OldNestsForNestedLogit,
    choice: ExpressionOrNumeric,
) -> Expression:
    """Implements the log of a nested logit model as a MEV model.

    :param util: dict of objects representing the utility functions of
        each alternative, indexed by numerical ids.

    :param availability: dict of objects representing the availability of each
        alternative (:math:`a_i` in the above formula), indexed
        by numerical ids. Must be consistent with util, or
        None. In this case, all alternatives are supposed to be
        always available.

    :param nests: object describing the nesting structure

    :param choice: id of the alternative for which the probability must be
              calculated.

    :return: log of choice probability for the nested logit model,
             based on the derivatives of the MEV generating function produced
             by the function getMevForNested

    :raise BiogemeError: if the definition of the nests is invalid.
    """
    if not isinstance(nests, NestsForNestedLogit):
        logger.warning(
            'It is recommended to define the nests of the nested logit model using '
            'the objects OneNestForNestedLogit and NestsForNestedLogit defined '
            'in biogeme.nests.'
        )
        nests = NestsForNestedLogit(choice_set=list(util), tuple_of_nests=nests)

    ok, message = nests.check_partition()
    if not ok:
        raise excep.BiogemeError(message)
    log_gi = get_mev_for_nested(
        util,
        availability,
        nests,
    )
    log_p = logmev(util, log_gi, availability, choice)
    return log_p


def nested_mev_mu(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForNestedLogit | OldNestsForNestedLogit,
    choice: ExpressionOrNumeric,
    mu: ExpressionOrNumeric,
) -> Expression:
    """Implements the nested logit model as a MEV model, where mu is also
    a parameter, if the user wants to test different normalization
    schemes.

    :param util: dict of objects representing the utility functions of
              each alternative, indexed by numerical ids.

    :param availability: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :param nests: object describing the nesting structure

    :param choice: id of the alternative for which the probability must be
              calculated.

    :param mu: expression producing the value of the top-level scale parameter.

    :return: the nested logit choice probability based on the following
             derivatives of the MEV generating function:

    .. math:: \\frac{\\partial G}{\\partial y_i}(e^{V_1},\\ldots,e^{V_J}) =
              \\mu e^{(\\mu_m-1)V_i} \\left(\\sum_{i=1}^{J_m}
              e^{\\mu_m V_i}\\right)^{\\frac{\\mu}{\\mu_m}-1}

    Where :math:`m` is the (only) nest containing alternative :math:`i`, and
    :math:`G` is the MEV generating function.


    """
    return exp(lognested_mev_mu(util, availability, nests, choice, mu))


@deprecated(nested_mev_mu)
def nestedMevMu(
    util: dict[int, Expression],
    availability: dict[int, Expression] | None,
    nests: NestsForNestedLogit | OldNestsForNestedLogit,
    choice: Expression,
    mu: Expression,
) -> Expression:
    pass


def lognested_mev_mu(
    util: dict[int, ExpressionOrNumeric],
    availability: dict[int, ExpressionOrNumeric] | None,
    nests: NestsForNestedLogit | OldNestsForNestedLogit,
    choice: ExpressionOrNumeric,
    mu: ExpressionOrNumeric,
) -> Expression:
    """Implements the log of the nested logit model as a MEV model, where
    mu is also a parameter, if the user wants to test different
    normalization schemes.


    :param util: dict of objects representing the utility functions of
        each alternative, indexed by numerical ids.

    :param availability: dict of objects representing the availability of each
               alternative (:math:`a_i` in the above formula), indexed
               by numerical ids. Must be consistent with util, or
               None. In this case, all alternatives are supposed to be
               always available.

    :param nests: object describing the nesting structure

    :param choice: id of the alternative for which the probability must be
              calculated.

    :param mu: expression producing the value of the top-level scale parameter.

    :return: the log of the nested logit choice probability based on the
        following derivatives of the MEV generating function:

        .. math:: \\frac{\\partial G}{\\partial y_i}(e^{V_1},\\ldots,e^{V_J}) =
                  \\mu e^{(\\mu_m-1)V_i} \\left(\\sum_{i=1}^{J_m}
                  e^{\\mu_m V_i}\\right)^{\\frac{\\mu}{\\mu_m}-1}

        where :math:`m` is the (only) nest containing alternative :math:`i`,
        and :math:`G` is the MEV generating function.


    """

    log_gi = get_mev_for_nested_mu(util, availability, nests, mu)
    log_p = logmev(util, log_gi, availability, choice)
    return log_p


@deprecated(lognested_mev_mu)
def lognestedMevMu(
    util: dict[int, Expression],
    availability: dict[int, Expression] | None,
    nests: NestsForNestedLogit | OldNestsForNestedLogit,
    choice: Expression,
    mu: Expression,
) -> Expression:
    pass



"""
.. _plot_b05normal_mixture:

Mixture of logit models
=======================

Example of a normal mixture of logit models, using Monte-Carlo integration.

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 17:30:14 2023
"""

import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, bioDraws, log, MonteCarlo

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.DEBUG)
logger.info('Example b05normal_mixtures.py')

# %%
# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation.
B_TIME = Beta('B_TIME', 0, None, None, 0)

# %%
# It is advised *not* to use 0 as starting value for the following parameter.
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL')

# %%
# Definition of the utility functions.
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Conditional to B_TIME_RND, we have a logit model (called the kernel).
prob = models.logit(V, av, CHOICE)

# %%
# We integrate over B_TIME_RND using Monte-Carlo.
logprob = log(MonteCarlo(prob))

# %%
# These notes will be included as such in the report file.
USER_NOTES = (
    'Example of a mixture of logit models with three alternatives, '
    'approximated using Monte-Carlo integration.'
)

# %%
# Create the Biogeme object. As the objective is to illustrate the
# syntax, we calculate the Monte-Carlo approximation with a small
# number of draws. To achieve that, we provide a parameter file
# different from the default one.
the_biogeme = bio.BIOGEME(
    database, logprob, userNotes=USER_NOTES, parameter_file='few_draws.toml'
)
the_biogeme.modelName = 'b05normal_mixture'

# %%
print(f'Number of draws: {the_biogeme.number_of_draws}')

# %%
# Estimate the parameters
results = the_biogeme.estimate()
# %%
print(results.short_summary())
# %%
pandas_results = results.getEstimatedParameters()
pandas_results



import sys
from collections import defaultdict
from itertools import product
from typing import Iterable, Iterator


class ModelNames:
    """Class generating model names from unique configuration string"""

    def __init__(self, prefix: str = 'Model'):
        self.prefix = prefix
        self.dict_of_names = {}
        self.current_number = 0

    def __call__(self, the_id: str) -> str:
        """Get a short model name from a unique ID

        :param the_id: id of the model
        :type the_id: str (or anything that can be used as a key for a dict)
        """
        the_name = self.dict_of_names.get(the_id)
        if the_name is None:
            the_name = f'{self.prefix}_{self.current_number:06d}'
            self.current_number += 1
            self.dict_of_names[the_id] = the_name
        return the_name


def generate_unique_ids(list_of_ids: list[str]) -> dict[str, str]:
    """If there are duplicates in the list, a new list is generated
    where there are renamed to obtain a list with unique IDs.

    :param list_of_ids: list of ids
    :type list_of_ids: list[str]

    :return: a dict that maps the unique names with the original name
    """
    counts = defaultdict(int)
    for the_id in list_of_ids:
        counts[the_id] += 1

    results = {}
    for name, count in counts.items():
        if count == 1:
            results[name] = name
        else:
            substitutes = [f'{name}_{i}' for i in range(count)]
            for new_name in substitutes:
                results[new_name] = name
    return results


def unique_product(*iterables: Iterable, max_memory_mb: int = 1024) -> Iterator[tuple]:
    """Generate the Cartesian product of multiple iterables, keeping
    only the unique entries.  Raises a MemoryError if memory usage
    exceeds the specified threshold.

    :param iterables: Variable number of iterables to compute the
        Cartesian product from.
    :type iterables: Iterable

    :param max_memory_mb: Maximum memory usage in megabytes (default: 1024MB).
    :type max_memory_mb: int

    :return: Yields unique entries from the Cartesian product.
    :rtype: Iterator[tuple]

    """

    mb_to_bytes = 1024 * 1024
    max_memory_bytes = max_memory_mb * mb_to_bytes  # Convert MB to bytes
    seen = set()  # Set to store seen entries
    total_memory = 0  # Track memory usage

    for items in product(*iterables):
        if items not in seen:
            seen.add(items)
            item_size = sum(sys.getsizeof(item) for item in items)
            total_memory += item_size
            if total_memory > max_memory_bytes:
                raise MemoryError(
                    f'Memory usage exceeded the specified threshold: '
                    f'{total_memory/mb_to_bytes:.1f} MB > '
                    f'{max_memory_bytes/mb_to_bytes} MB.'
                )
            yield items



import pickle
import subprocess
import sys

import jax

from biogeme.tools import TemporaryFile


def evaluate_hessian_in_subprocess(free_betas_values, data, draws, rv, sum_function):
    """
    Launch a subprocess to compute the Hessian using JAX autodiff.
    """

    import os

    with TemporaryFile() as tmp_file:
        hess_file = tmp_file.fullpath
        print(f"[DEBUG] Created temporary file: {hess_file}")

        args = (free_betas_values, data, draws, rv, sum_function, hess_file)
        payload = pickle.dumps(args)
        cmd = [sys.executable, __file__]
        try:
            result = subprocess.run(
                cmd, input=payload, stderr=subprocess.PIPE, timeout=60
            )
            print(f"[DEBUG] Subprocess return code: {result.returncode}")
            if result.stderr:
                print(f"[DEBUG] Subprocess stderr:\n{result.stderr.decode()}")

            if result.returncode == 0:
                if not os.path.exists(hess_file):
                    print(f"[DEBUG] File does not exist: {hess_file}")
                    return None
                try:
                    with open(hess_file, "rb") as f:
                        print(f"[DEBUG] Reading result from: {hess_file}")
                        return pickle.load(f)
                except (
                    pickle.UnpicklingError,
                    EOFError,
                    AttributeError,
                    FileNotFoundError,
                ) as e:
                    print(f"[DEBUG] Failed to unpickle result: {e}")
                    return None
        finally:
            try:
                os.remove(hess_file)
                print(f"[DEBUG] Deleted temporary file: {hess_file}")
            except FileNotFoundError:
                print(f"[DEBUG] Could not delete missing file: {hess_file}")
        return None


if __name__ == "__main__":
    free_betas_values, data, draws, rv, sum_function, hess_file = pickle.loads(
        sys.stdin.buffer.read()
    )
    print(f"[DEBUG subprocess] Writing result to: {hess_file}")
    fn = lambda p, d, r, rv: sum_function(p, d, r, rv)[0]
    jitted_fn = jax.jit(fn)
    hessian = jax.jacfwd(jax.grad(jitted_fn, argnums=0), argnums=0)(
        free_betas_values, data, draws, rv
    )
    try:
        with open(hess_file, "wb") as f:
            pickle.dump(hessian, f)
            print(f"[DEBUG subprocess] Successfully wrote Hessian to: {hess_file}")
    except Exception as e:
        print(f"[DEBUG subprocess] Failed to write Hessian: {e}")



"""Module to create callable functions from Biogeme expressions.

This utility allows converting a symbolic Biogeme expression into a callable
function that can be evaluated with a given parameter vector.

Michel Bierlaire
Fri Mar 28 19:03:28 2025
"""

from __future__ import annotations

from typing import Protocol

import numpy as np

from biogeme.function_output import FunctionOutput, NamedFunctionOutput
from .single_formula import CompiledFormulaEvaluator
from ..constants import LOG_LIKE
from ..database import Database
from ..expressions import Expression
from ..model_elements import ModelElements
from ..second_derivatives import SecondDerivativesMode


class CallableExpression(Protocol):
    def __call__(
        self,
        x: np.ndarray,
        gradient: bool,
        hessian: bool,
        bhhh: bool,
    ) -> FunctionOutput: ...


class NamedCallableExpression(Protocol):
    def __call__(
        self,
        x: np.ndarray,
        gradient: bool,
        hessian: bool,
        bhhh: bool,
    ) -> NamedFunctionOutput: ...


def function_from_compiled_formula(
    the_compiled_function: CompiledFormulaEvaluator,
    the_betas: dict[str, float],
    named_output: bool = False,
) -> CallableExpression | NamedCallableExpression:
    """Create a callable function from a symbolic Biogeme expression.

    :param the_compiled_function: Compiled function evaluator.
    :param the_betas: Dictionary mapping parameter names to initial values.
    :param named_output: if True, the entries of the derivatives are associated with parameter names

    :return: A callable that takes a NumPy array of parameter values
             and returns a FunctionOutput.
    """

    if named_output:

        def the_named_function(
            x: np.ndarray,
            gradient: bool,
            hessian: bool,
            bhhh: bool,
        ) -> NamedFunctionOutput:
            """Evaluate the Biogeme expression with updated parameter values.

            :param x: A NumPy array of new parameter values.
            :param gradient: If True, compute the gradient of the function.
            :param hessian: If True, compute the Hessian of the function.
            :param bhhh: If True, compute the BHHH matrix (outer product of gradients).
            :return: The evaluated FunctionOutput object.
            """
            the_betas.update(dict(zip(the_betas.keys(), x)))
            result: FunctionOutput = the_compiled_function.evaluate(
                the_betas=the_betas, gradient=gradient, hessian=hessian, bhhh=bhhh
            )
            named_result = the_compiled_function.model_elements.generate_named_output(
                function_output=result
            )
            return named_result

        return the_named_function

    def the_function(
        x: np.ndarray,
        gradient: bool,
        hessian: bool,
        bhhh: bool,
    ) -> FunctionOutput:
        """Evaluate the Biogeme expression with updated parameter values.

        :param x: A NumPy array of new parameter values.
        :param gradient: If True, compute the gradient of the function.
        :param hessian: If True, compute the Hessian of the function.
        :param bhhh: If True, compute the BHHH matrix (outer product of gradients).
        :return: The evaluated FunctionOutput object.
        """
        the_betas.update(dict(zip(the_betas.keys(), x)))
        return the_compiled_function.evaluate(
            the_betas=the_betas, gradient=gradient, hessian=hessian, bhhh=bhhh
        )

    return the_function


def function_from_expression(
    expression: Expression,
    database: Database,
    numerically_safe: bool,
    the_betas: dict[str, float],
    number_of_draws: int | None = None,
    named_output: bool = False,
) -> CallableExpression | NamedCallableExpression:

    model_elements = ModelElements(
        expressions={LOG_LIKE: expression},
        database=database,
        number_of_draws=number_of_draws,
    )
    compiled_formula = CompiledFormulaEvaluator(
        model_elements=model_elements,
        second_derivatives_mode=SecondDerivativesMode.ANALYTICAL,
        numerically_safe=numerically_safe,
    )
    the_function = function_from_compiled_formula(
        the_compiled_function=compiled_formula,
        the_betas=the_betas,
        named_output=named_output,
    )
    return the_function



from .segmentation_context import DiscreteSegmentationTuple
from .segmentation import Segmentation
from .segmented_beta import segmented_beta
from .database import verify_segmentation, generate_segmentation



"""Implementation of the "generalized translated utility function" MDCEV model. See the technical report.

Michel Bierlaire
Tue Apr 9 08:44:12 2024
"""

from typing import Callable

import numpy as np

from biogeme.database import Database
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Expression, log, Numeric, exp
from biogeme.mdcev.mdcev import Mdcev, MdcevConfiguration
from biogeme.tools.checks import validate_dict_types


class Generalized(Mdcev):
    """Implementation of the MDCEV model with generalized translated utility function."""

    def __init__(
        self,
        model_name: str,
        baseline_utilities: dict[int, Expression],
        gamma_parameters: dict[int, Expression | None],
        alpha_parameters: dict[int, Expression] | None = None,
        scale_parameter: Expression | None = None,
        prices: dict[int, Expression] | None = None,
        weights: Expression | None = None,
    ) -> None:
        """Implementation of the MDCEV model with Gamma profile."""
        super().__init__(
            model_name=model_name,
            baseline_utilities=baseline_utilities,
            gamma_parameters=gamma_parameters,
            alpha_parameters=alpha_parameters,
            scale_parameter=scale_parameter,
            weights=weights,
        )
        self.prices: dict[int, Expression] | None = prices
        if self.prices is not None:
            validate_dict_types(self.prices, 'prices', Expression)

    def transformed_utility(
        self,
        the_id: int,
        the_consumption: Expression,
    ) -> Expression:
        """Calculates the utility for one alternative."""

        if the_id not in self.alternatives:
            error_msg = (
                f'Alternative id {the_id} is invalid. Valid ids: {self.alternatives}'
            )
            raise BiogemeError(error_msg)
        gamma: Expression | None = self.gamma_parameters[the_id]
        price: Expression | None = None if self.prices is None else self.prices[the_id]

        baseline_utility = self.baseline_utilities[the_id]
        alpha_parameter = self.alpha_parameters[the_id]

        def no_gamma_no_price() -> Expression:
            """gamma is None. price is None"""
            return baseline_utility + (alpha_parameter - 1) * log(the_consumption)

        def no_gamma_yes_price() -> Expression:
            """gamma is None. price is not None"""
            return (
                baseline_utility
                + (alpha_parameter - 1) * log(the_consumption)
                - alpha_parameter * log(price)
            )

        def yes_gamma_no_price() -> Expression:
            """gamma is not None. price is None"""
            return baseline_utility + (alpha_parameter - 1) * log(
                1.0 + the_consumption / gamma
            )

        def yes_gamma_yes_price() -> Expression:
            """gamma is not None. price is not None"""
            return (
                baseline_utility
                - log(price)
                + (alpha_parameter - 1) * log(1.0 + the_consumption / (price * gamma))
            )

        dict_of_configurations: dict[
            MdcevConfiguration, Callable[[], Expression | float]
        ] = {
            MdcevConfiguration(
                gamma_is_none=True, price_is_none=True
            ): no_gamma_no_price,
            MdcevConfiguration(
                gamma_is_none=True, price_is_none=False
            ): no_gamma_yes_price,
            MdcevConfiguration(
                gamma_is_none=False, price_is_none=True
            ): yes_gamma_no_price,
            MdcevConfiguration(
                gamma_is_none=False, price_is_none=False
            ): yes_gamma_yes_price,
        }

        current_configuration = MdcevConfiguration(
            gamma_is_none=gamma is None, price_is_none=price is None
        )

        return dict_of_configurations[current_configuration]()

    def calculate_log_determinant_one_alternative(
        self, the_id: int, consumption: Expression
    ) -> Expression:
        """Calculate the log of the entries for the determinant. For
            the outside good, gamma is equal to 0.

        :param the_id: identifier of the good.
        :param consumption: expression for the consumption.
        """
        if the_id not in self.alternatives:
            error_msg = (
                f'Alternative id {the_id} is invalid. Valid ids: {self.alternatives}'
            )
            raise BiogemeError(error_msg)
        price = None if self.prices is None else self.prices[the_id]
        gamma = self.gamma_parameters[the_id]
        if gamma is None:
            return log(Numeric(1) - self.alpha_parameters[the_id]) - log(consumption)
        if price is None:
            return log(Numeric(1) - self.alpha_parameters[the_id]) - log(
                consumption + gamma
            )
        return log(Numeric(1) - self.alpha_parameters[the_id]) - log(
            consumption + price * gamma
        )

    def calculate_inverse_of_determinant_one_alternative(
        self, the_id: int, consumption: Expression
    ) -> Expression:
        """Calculate the inverse of the entries for the determinant. For
            the outside good, gamma is equal to 0.

        :param the_id: identifier of the good.
        :param consumption: expression for the consumption.
        """
        if the_id not in self.alternatives:
            error_msg = (
                f'Alternative id {the_id} is invalid. Valid ids: {self.alternatives}'
            )
            raise BiogemeError(error_msg)
        price = None if self.prices is None else self.prices[the_id]
        gamma = self.gamma_parameters[the_id]
        if gamma is None:
            return consumption / (Numeric(1) - self.alpha_parameters[the_id])
        if price is None:
            return (consumption + gamma) / (Numeric(1) - self.alpha_parameters[the_id])
        return (consumption + price * gamma) / (
            Numeric(1) - self.alpha_parameters[the_id]
        )

    def utility_expression_one_alternative(
        self,
        the_id: int,
        the_consumption: Expression,
        unscaled_epsilon: Expression,
    ) -> Expression:
        """Utility expression. Used only for code validation."""
        baseline_utility = self.baseline_utilities[the_id]
        epsilon: Expression = (
            unscaled_epsilon
            if self.scale_parameter is None
            else unscaled_epsilon / self.scale_parameter
        )
        price: Expression = Numeric(1) if self.prices is None else self.prices[the_id]
        alpha: Expression = self.alpha_parameters[the_id]
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            return (
                exp(baseline_utility + epsilon)
                * (the_consumption / price) ** alpha
                / alpha
            )
        the_term = (1 + the_consumption / (price * gamma)) ** alpha
        return exp(baseline_utility + epsilon) * gamma * (the_term - 1) / alpha

    def utility_one_alternative(
        self,
        the_id: int,
        the_consumption: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Utility needed for forecasting"""
        baseline_utility = self.calculate_baseline_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        if self.scale_parameter is not None:
            epsilon /= self.scale_parameter.get_value()
        price: float = 1 if self.prices is None else self.prices[the_id].get_value()
        alpha: float = self.alpha_parameters[the_id].get_value()
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            return (
                np.exp(baseline_utility + epsilon)
                * (the_consumption / price) ** alpha
                / alpha
            )
        the_term = (1 + the_consumption / (price * gamma.get_value())) ** alpha
        return (
            np.exp(baseline_utility + epsilon)
            * gamma.get_value()
            * (the_term - 1)
            / alpha
        )

    def sorting_utility_one_alternative(
        self,
        alternative_id: int,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Utility used to sort the alternatives. Used in the forecasting algorithm to identify chosen and non chpsen
        alternatives"""
        baseline_utility = self.calculate_baseline_utility(
            alternative_id=alternative_id, one_observation=one_observation
        )
        if self.scale_parameter is not None:
            epsilon /= self.scale_parameter.get_value()

        if self.prices:
            price = self.prices[alternative_id].get_value()
            return baseline_utility + epsilon - np.log(price)
        return baseline_utility + epsilon

    def derivative_utility_one_alternative(
        self,
        the_id: int,
        the_consumption: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Used in the optimization problem solved for forecasting tp calculate the dual variable."""
        baseline_utility = self.calculate_baseline_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        if self.scale_parameter is not None:
            epsilon /= self.scale_parameter.get_value()
        price: float = 1.0 if self.prices is None else self.prices[the_id].get_value()
        alpha: float = self.alpha_parameters[the_id].get_value()
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            return (
                np.exp(baseline_utility + epsilon)
                * (the_consumption / price) ** (alpha - 1)
                / price
            )
        return (
            np.exp(baseline_utility + epsilon)
            * (1 + the_consumption / (price * gamma.get_value())) ** (alpha - 1)
            / price
        )

    def optimal_consumption_one_alternative(
        self,
        the_id: int,
        dual_variable: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Analytical calculation of the optimal consumption if the dual variable is known."""
        baseline_utility = self.calculate_baseline_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        if self.scale_parameter is not None:
            epsilon /= self.scale_parameter.get_value()
        price: float = 1.0 if self.prices is None else self.prices[the_id].get_value()
        alpha: float = self.alpha_parameters[the_id].get_value()
        gamma: Expression | None = self.gamma_parameters[the_id]
        numerator = price * dual_variable
        denominator = np.exp(baseline_utility + epsilon)
        ratio = numerator / denominator
        exponent = 1 / (alpha - 1)
        if gamma is None:
            return price * ratio**exponent
        return price * gamma.get_value() * (ratio**exponent - 1)

    def lower_bound_dual_variable(
        self,
        chosen_alternatives: set[int],
        one_observation: Database,
        epsilon: np.ndarray,
    ) -> float:
        """Method providing model specific bounds on the dual variable. It not overloaded,
        default values are used.

        :param chosen_alternatives: list of alternatives that are chosen at the optimal solution
        :param one_observation: data for one observation.
        :param epsilon: draws from the error term.
        :return: a lower bound on the dual variable, such that the expenditure calculated for any larger value is
        well-defined and non negative.
        """
        return 0.0

    def _list_of_expressions(self) -> list[Expression]:
        """Extract the list of expressions involved in the model"""
        if self.prices is None:
            return []
        return [expression for expression in self.prices.values()]



"""Tests for unary expressions

Michel Bierlaire
Wed Mar 26 08:35:13 2025
"""

import unittest
import warnings

import jax.numpy as jnp
import numpy as np

from biogeme.exceptions import BiogemeError
from biogeme.expressions import (
    BelongsTo,
    Derive,
    Expression,
    IntegrateNormal,
    MonteCarlo,
    NormalCdf,
    PanelLikelihoodTrajectory,
    bioNormalCdf,
    cos,
    exp,
    log,
    logzero,
    sin,
)
from biogeme.expressions.power_constant import PowerConstant
from biogeme.expressions.unary_minus import UnaryMinus


class DummyNumeric(Expression):
    def __init__(self, value: float):
        super().__init__()
        self.value = value

    def get_value(self) -> float:
        return self.value

    def recursive_construct_jax_function(self, numerically_safe: bool):
        def f(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            draws: jnp.ndarray,
            random_variables: jnp.ndarray,
        ) -> float:
            return self.value

        return f

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return repr(self.value)


class TestUnaryExpressions(unittest.TestCase):
    def setUp(self):
        self.constant = DummyNumeric(2.0)
        self.zero = DummyNumeric(0.0)
        self.negative = DummyNumeric(-4.0)

    def test_unary_minus(self):
        expr = UnaryMinus(self.constant)
        self.assertEqual(expr.get_value(), -2.0)
        self.assertEqual(str(expr), f'(-{self.constant})')
        self.assertIn('UnaryMinus', repr(expr))

    def test_exp(self):
        expr = exp(self.constant)
        self.assertAlmostEqual(expr.get_value(), np.exp(2.0))
        self.assertIn('exp', repr(expr))

    def test_log(self):
        expr = log(self.constant)
        self.assertAlmostEqual(expr.get_value(), np.log(2.0))
        self.assertIn('log', repr(expr))

    def test_logzero_nonzero(self):
        expr = logzero(self.constant)
        self.assertAlmostEqual(expr.get_value(), np.log(2.0))

    def test_logzero_zero(self):
        expr = logzero(self.zero)
        self.assertEqual(expr.get_value(), 0.0)

    def test_sin(self):
        expr = sin(self.constant)
        self.assertAlmostEqual(expr.get_value(), np.sin(2.0))

    def test_cos(self):
        expr = cos(self.constant)
        self.assertAlmostEqual(expr.get_value(), np.cos(2.0))

    def test_power_constant_positive(self):
        expr = PowerConstant(self.constant, 2)
        self.assertEqual(expr.get_value(), 4.0)

    def test_power_constant_negative_integer(self):
        expr = PowerConstant(self.negative, 2)
        self.assertEqual(expr.get_value(), 16.0)

    def test_power_constant_negative_nonint(self):
        expr = PowerConstant(self.negative, 2.5)
        with self.assertRaises(BiogemeError):
            expr.get_value()

    def test_panel_likelihood_repr(self):
        expr = PanelLikelihoodTrajectory(self.constant)
        self.assertIn('PanelLikelihoodTrajectory', repr(expr))

    def test_normal_cdf_repr(self):
        expr = NormalCdf(self.constant)
        print(repr(expr))
        self.assertIn('NormalCdf', repr(expr))

    def test_monte_carlo_jax(self):
        expr = MonteCarlo(self.constant)
        jax_fn = expr.recursive_construct_jax_function(numerically_safe=False)
        draws = jnp.array([[1.0], [2.0], [3.0]])
        self.assertAlmostEqual(jax_fn(None, None, draws, None), 2.0)

    def test_draws_deprecation_warning(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            _ = bioNormalCdf(self.constant)
            self.assertTrue(
                any(issubclass(wi.category, DeprecationWarning) for wi in w)
            )

    def test_derive_repr(self):
        expr = Derive(self.constant, "beta")
        self.assertIn("Derive", repr(expr))

    def test_integrate_repr(self):
        expr = IntegrateNormal(self.constant, "z")
        self.assertIn("Integrate", repr(expr))

    def test_belongsto_repr(self):
        expr = BelongsTo(self.constant, {1.0, 2.0})
        self.assertIn("BelongsTo", repr(expr))


if __name__ == '__main__':
    unittest.main()



"""

biogeme.nests
=============

Examples of use of several functions.

This is designed for programmers who need examples of use of the
functions of the module. The examples are designed to illustrate the
syntax. They do not correspond to any meaningful model.

:author: Michel Bierlaire
:date: Wed Nov 29 18:35:06 2023
"""
import numpy as np
from biogeme.version import getText
from biogeme.nests import (
    OneNestForNestedLogit,
    OneNestForCrossNestedLogit,
    NestsForCrossNestedLogit,
    NestsForNestedLogit,
)


# %%
# Version of Biogeme.
print(getText())


# %%
# Covariance and correlation between two alternatives of a
# cross-nested logit model. Here, we test a logit model by setting the
# nest parameters to 1.0. We expect the identify matrix as correlation.
choice_set = ['i', 'j', 'k', 'm']
mu_nest_1 = 1.0
alphas_1 = {'i': 1, 'j': 1}
nest_1 = OneNestForCrossNestedLogit(
    nest_param=mu_nest_1, dict_of_alpha=alphas_1, name='Nest 1'
)
mu_nest_2 = 1.0
alphas_2 = {'j': 0.0, 'k': 1, 'm': 1}
nest_2 = OneNestForCrossNestedLogit(
    nest_param=mu_nest_2, dict_of_alpha=alphas_2, name='Nest 2'
)
nests = NestsForCrossNestedLogit(choice_set=choice_set, tuple_of_nests=(nest_1, nest_2))


# %%
nests.correlation()

# %%
# Entries of the covariance matrix can also be obtained. Here, we
# report the variance for alternative `i`.
nests.covariance('i', 'i')

# %%
# It is :math:`\pi^2/6`.
np.pi**2 / 6

# %%
# Second, a nested logit model
mu_nest_1 = 1.5
alphas_1 = {'i': 1, 'j': 1}
nest_1 = OneNestForNestedLogit(
    nest_param=mu_nest_1, list_of_alternatives=['i', 'j'], name='Nest 1'
)
mu_nest_2 = 2.0
alphas_2 = {'k': 1, 'm': 1}
nest_2 = OneNestForNestedLogit(
    nest_param=mu_nest_2, list_of_alternatives=['k', 'm'], name='Nest 2'
)
nests = NestsForNestedLogit(choice_set=choice_set, tuple_of_nests=(nest_1, nest_2))

# %%
nests.correlation()

# %%
# Theoretical value for the correlation
correl_nest_1 = 1 - (1 / mu_nest_1**2)
correl_nest_1

# %%
correl_nest_2 = 1 - (1 / mu_nest_2**2)
correl_nest_2


# %%
# The same nested logit model, coded as a cross-nested logit


mu_nest_1 = 1.5
alphas_1 = {'i': 1, 'j': 1}
nest_1 = OneNestForCrossNestedLogit(
    nest_param=mu_nest_1, dict_of_alpha=alphas_1, name='Nest 1'
)
mu_nest_2 = 2.0
alphas_2 = {'j': 0.0, 'k': 1, 'm': 1}
nest_2 = OneNestForCrossNestedLogit(
    nest_param=mu_nest_2, dict_of_alpha=alphas_2, name='Nest 2'
)
nests = NestsForCrossNestedLogit(choice_set=choice_set, tuple_of_nests=(nest_1, nest_2))


# %%
nests.correlation()

# %%
# Finally, a cross-nested logit model, where alternative j is
# correlated with all the other alternatives, and belong to two
# different nests.
mu_nest_1 = 1.5
alphas_1 = {'i': 1, 'j': 0.5}
nest_1 = OneNestForCrossNestedLogit(
    nest_param=mu_nest_1, dict_of_alpha=alphas_1, name='Nest 1'
)
mu_nest_2 = 2.0
alphas_2 = {'j': 0.5, 'k': 1, 'm': 1}
nest_2 = OneNestForCrossNestedLogit(
    nest_param=mu_nest_2, dict_of_alpha=alphas_2, name='Nest 2'
)
nests = NestsForCrossNestedLogit(choice_set=choice_set, tuple_of_nests=(nest_1, nest_2))

# %%
nests.correlation()



"""

biogeme.database
================

Examples of use of several functions.

This is designed for programmers who need examples of use of the
functions of the module. The examples are designed to illustrate the
syntax. They do not correspond to any meaningful model.

:author: Michel Bierlaire
:date: Thu Nov 16 18:36:59 2023
"""

import biogeme.version as ver
import pandas as pd
import numpy as np
import biogeme.database as db
from biogeme.expressions import Variable, exp, bioDraws
from biogeme.expressions import TypeOfElementaryExpression
from biogeme.segmentation import DiscreteSegmentationTuple
from biogeme.exceptions import BiogemeError

# %%
# Version of Biogeme.
print(ver.getText())

# %%
# We set the seed so that the outcome of random operations is always the same.
np.random.seed(90267)

# %%
# Create a database from a pandas data frame.
df = pd.DataFrame(
    {
        'Person': [1, 1, 1, 2, 2],
        'Exclude': [0, 0, 1, 0, 1],
        'Variable1': [1, 2, 3, 4, 5],
        'Variable2': [10, 20, 30, 40, 50],
        'Choice': [1, 2, 3, 1, 2],
        'Av1': [0, 1, 1, 1, 1],
        'Av2': [1, 1, 1, 1, 1],
        'Av3': [0, 1, 1, 1, 1],
    }
)
my_data = db.Database('test', df)
print(my_data)

# %%
# `valuesFromDatabase`: evaluates an expression for each entry of the
# database. Takes as argument an expression, and returns a numpy
# series, long as the number of entries in the database, containing
# the calculated quantities.

# %%
Variable1 = Variable('Variable1')
Variable2 = Variable('Variable2')
expr = Variable1 + Variable2
result = my_data.valuesFromDatabase(expr)
print(result)

# %%
# `check_segmentation`: checks that the segmentation covers the complete database.
# A segmentation is a partition of the dataset based on the value of
# one of the variables. For instance, we can segment on the Choice
# variable.

# %%
correct_mapping = {1: 'Alt. 1', 2: 'Alt. 2', 3: 'Alt. 3'}
correct_segmentation = DiscreteSegmentationTuple(
    variable='Choice', mapping=correct_mapping
)

# %%
# If the segmentation is well defined, the function returns the size
# of each segment in the database.

# %%
my_data.check_segmentation(correct_segmentation)

# %%
incorrect_mapping = {1: 'Alt. 1', 2: 'Alt. 2'}
incorrect_segmentation = DiscreteSegmentationTuple(
    variable='Choice', mapping=incorrect_mapping
)

# %%
# If the segmentation is incorrect, an exception is raised.

# %%
try:
    my_data.check_segmentation(incorrect_segmentation)
except BiogemeError as e:
    print(e)

# %%
another_incorrect_mapping = {1: 'Alt. 1', 2: 'Alt. 2', 4: 'Does not exist'}
another_incorrect_segmentation = DiscreteSegmentationTuple(
    variable='Choice', mapping=another_incorrect_mapping
)

# %%
try:
    my_data.check_segmentation(another_incorrect_segmentation)
except BiogemeError as e:
    print(e)

# %%
# `checkAvailabilityOfChosenAlt`: check if the chosen alternative
# is available for each entry in the database.
# %%
Av1 = Variable('Av1')
Av2 = Variable('Av2')
Av3 = Variable('Av3')
Choice = Variable('Choice')
avail = {1: Av1, 2: Av2, 3: Av3}
result = my_data.checkAvailabilityOfChosenAlt(avail, Choice)
print(result)

# %%
# `choiceAvailabilityStatistics`: calculates the number of time an
# alternative is chosen and available.
my_data.choiceAvailabilityStatistics(avail, Choice)

# %%
# Suggest a scaling of the variables in the database
# %%
my_data.data.columns

# %%
my_data.suggestScaling()

# %%
my_data.suggestScaling(columns=['Variable1', 'Variable2'])

# %%
# `scaleColumn`: divide an entire column by a scale value
# %%
# Before.
my_data.data

# %%
my_data.scaleColumn('Variable2', 0.01)

# %%
# After.
my_data.data

# %%
# `addColumn`: add a new column in the database, calculated from an expression.
# %%
Variable1 = Variable('Variable1')
Variable2 = Variable('Variable2')
expression = exp(0.5 * Variable2) / Variable1
expression = Variable2 * Variable1
result = my_data.addColumn(expression, 'NewVariable')
print(my_data.data['NewVariable'].tolist())

# %%
my_data.data

# %%
# `split`: shuffle the data, and split the data into slices. For each
# slide, an estimation and a validation sets are generated. The
# validation set is the slice itself. The estimation set is the rest
# of the data.

# %%
dataSets = my_data.split(3)
for i in dataSets:
    print("==========")
    print("Estimation:")
    print(type(i[0]))
    print(i[0])
    print("Validation:")
    print(i[1])

# %%
# `count`: counts the number of observations that have a specific
# value in a given column.

# %%
# For instance, counts the number of entries for individual 1.
# %%
my_data.count('Person', 1)

# %%
# `remove`: removes from the database all entries such that the value
# of the expression is not 0.
# %%
exclude = Variable('Exclude')
my_data.remove(exclude)
my_data.data

# %%
# `dumpOnFile`: dumps the database in a CSV formatted file.
my_data.dumpOnFile()

# %%
# %%bash
# cat test_dumped.dat

# %%

# `generateDraws`: generate draws for each variable. Takes as argument
#                  a dict indexed by the names of the variables,
#                  describing the types of draws. Each of them can be
#                  a native type or any type defined by the function
#                  database.setRandomNumberGenerators, as well as the
#                  list of names of the variables that require draws
#                  to be generated.  It returns a 3-dimensional table
#                  with draws. The 3 dimensions are
#
#               1. number of individuals
#               2. number of draws
#               3. number of variables

# %%
# List of native types and their description
db.Database.descriptionOfNativeDraws()

# %%
random_draws1 = bioDraws('random_draws1', 'NORMAL_MLHS_ANTI')
random_draws2 = bioDraws('random_draws2', 'UNIFORM_MLHS_ANTI')
random_draws3 = bioDraws('random_draws3', 'UNIFORMSYM_MLHS_ANTI')

# %%
# We build an expression that involves the three random variables
x = random_draws1 + random_draws2 + random_draws3
types = {
    key: the_draw.drawType
    for key, the_draw in x.dict_of_elementary_expression(
        TypeOfElementaryExpression.DRAWS
    ).items()
}
print(types)

# %%
# Generation of the draws.
the_draws_table = my_data.generateDraws(
    types, ['random_draws1', 'random_draws2', 'random_draws3'], 10
)
the_draws_table

# %%
# `setRandomNumberGenerators`: defines user-defined random numbers
# generators. It takes as arguments a dictionary of generators. The
# keys of the dictionary characterize the name of the generators, and
# must be different from the pre-defined generators in Biogeme:
# NORMAL, UNIFORM and UNIFORMSYM. The elements of the dictionary are
# functions that take two arguments: the number of series to generate
# (typically, the size of the database), and the number of draws per
# series.

# %%
# We first define functions returning draws, given the number of
# observations, and the number of draws


# %%
# A lognormal distribution.
def logNormalDraws(sample_size, number_of_draws):
    return np.exp(np.random.randn(sample_size, number_of_draws))


# %%
# An exponential distribution.
def exponentialDraws(sample_size, number_of_draws):
    return -1.0 * np.log(np.random.rand(sample_size, number_of_draws))


# %%
# We associate these functions with a name in a dictionary.
# %%
dict = {
    'LOGNORMAL': (logNormalDraws, 'Draws from lognormal distribution'),
    'EXP': (exponentialDraws, 'Draws from exponential distributions'),
}
my_data.setRandomNumberGenerators(dict)

# %%
# We can now generate draws from these distributions.
random_draws1 = bioDraws('random_draws1', 'LOGNORMAL')
random_draws2 = bioDraws('random_draws2', 'EXP')
x = random_draws1 + random_draws2
types = {
    key: the_draw.drawType
    for key, the_draw in x.dict_of_elementary_expression(
        TypeOfElementaryExpression.DRAWS
    ).items()
}
the_draws_table = my_data.generateDraws(types, ['random_draws1', 'random_draws2'], 10)
print(the_draws_table)

# %%
# `sampleWithReplacement`: extracts a random sample from the database,
# with replacement. Useful for bootstrapping.

# %%
my_data.sampleWithReplacement()

# %%
my_data.sampleWithReplacement(6)

# %%
# `panel`: defines the data as panel data. Takes as argument the name
# of the column that identifies individuals.
my_panel_data = db.Database('test', df)

# %%
# Data is not considered panel yet
my_panel_data.isPanel()

# %%
my_panel_data.panel('Person')

# %%
# Now it is panel.
print(my_panel_data.isPanel())

# %%
print(my_panel_data)

# %%
# When draws are generated for panel data, a set of draws is generated
# per person, not per observation.
random_draws1 = bioDraws('random_draws1', 'NORMAL')
random_draws2 = bioDraws('random_draws2', 'UNIFORM_HALTON3')

# %%
# We build an expression that involves the two random variables
x = random_draws1 + random_draws2
types = x.dict_of_elementary_expression(TypeOfElementaryExpression.DRAWS)
the_draws_table = my_panel_data.generateDraws(
    types, ['random_draws1', 'random_draws2'], 10
)
print(the_draws_table)

# %%
# `getNumberOfObservations`: reports the number of observations in the
# database. Note that it returns the same value, irrespectively if the
# database contains panel data or not.
my_data.getNumberOfObservations()

# %%
my_panel_data.getNumberOfObservations()

# %%
# `getSampleSize`: reports the size of the sample. If the data is
# cross-sectional, it is the number of observations in the
# database. If the data is panel, it is the number of individuals.
my_data.getSampleSize()

# %%
my_panel_data.getSampleSize()

# %%
# `sampleIndividualMapWithReplacement`: extracts a random sample of
# the individual map from a panel data database, with
# replacement. Useful for bootstrapping.
my_panel_data.sampleIndividualMapWithReplacement(10)



import unittest


from biogeme.catalog import (
    Controller,
    Catalog,
    CentralController,
    Configuration,
    SelectionTuple,
    ControllerOperator,
)
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Numeric, NamedExpression


class TestCentralController(unittest.TestCase):
    def setUp(self):

        expr_1 = Numeric(1)
        expr_2 = Numeric(2)
        expr_3 = Numeric(3)
        expr_4 = Numeric(4)
        expr_5 = Numeric(5)
        expr_6 = Numeric(6)
        expr_7 = Numeric(7)

        # Non nested catalogs
        named_expr_1 = NamedExpression(name='option1', expression=expr_1)
        named_expr_2 = NamedExpression(name='option2', expression=expr_2)
        test_controller_1 = Controller("c1", ["option1", "option2"])
        catalog_1 = Catalog(
            catalog_name='catalog_1',
            named_expressions=[named_expr_1, named_expr_2],
            controlled_by=test_controller_1,
        )
        named_expr_3 = NamedExpression(name='optionA', expression=expr_3)
        named_expr_4 = NamedExpression(name='optionB', expression=expr_4)
        named_expr_5 = NamedExpression(name='optionC', expression=expr_5)
        test_controller_2 = Controller("c2", ["optionA", "optionB", "optionC"])
        catalog_2 = Catalog(
            catalog_name='catalog_2',
            named_expressions=[named_expr_3, named_expr_4, named_expr_5],
            controlled_by=test_controller_2,
        )
        named_expr_6 = NamedExpression(name='choiceX', expression=expr_6)
        named_expr_7 = NamedExpression(name='choiceY', expression=expr_7)
        test_controller_3 = Controller("c3", ["choiceX", "choiceY"])
        catalog_3 = Catalog(
            catalog_name='catalog_3',
            named_expressions=[named_expr_6, named_expr_7],
            controlled_by=test_controller_3,
        )
        test_expression = catalog_1 + catalog_2 + catalog_3
        self.test_central_controller: CentralController = CentralController(
            expression=test_expression
        )

        # Nested catalogs
        named_expr_1 = NamedExpression(name='expr_1', expression=expr_1)
        named_expr_2 = NamedExpression(name='expr_2', expression=expr_2)
        named_expr_3 = NamedExpression(name='expr_3', expression=expr_3)
        named_expr_4 = NamedExpression(name='expr_4', expression=expr_4)
        controller_1 = Controller(
            controller_name='controller_1', specification_names=['expr_1', 'expr_2']
        )
        catalog_1 = Catalog(
            catalog_name='catalog_1',
            named_expressions=[named_expr_1, named_expr_2],
            controlled_by=controller_1,
        )
        named_catalog_1 = NamedExpression('catalog_1', catalog_1)
        controller_2 = Controller(
            controller_name='controller_2',
            specification_names=['expr_3', 'expr_4', 'catalog_1'],
        )
        catalog_2 = Catalog(
            catalog_name='catalog_2',
            named_expressions=[named_expr_3, named_expr_4, named_catalog_1],
            controlled_by=controller_2,
        )
        the_expression = Numeric(1) + catalog_2
        self.central_controller: CentralController = CentralController(
            expression=the_expression
        )
        self.configs = self.central_controller.all_configurations

    def test_number_of_configurations(self):
        # Add test cases for the number of configurations

        self.assertEqual(6, self.central_controller.number_of_configurations())

    def test_get_configuration(self):
        # Add test cases for get_configuration
        # Ensure the returned configuration matches the initial state
        expected_config = Configuration(
            [
                SelectionTuple(controller='controller_1', selection='expr_1'),
                SelectionTuple(controller='controller_2', selection='catalog_1'),
            ]
        )
        self.central_controller.set_configuration(expected_config)
        self.assertEqual(self.central_controller.get_configuration(), expected_config)

    def test_set_configuration(self):
        # Add test cases for set_configuration
        # Ensure it correctly sets the controller state
        config_to_set = Configuration(
            [
                SelectionTuple(controller='controller_1', selection='expr_1'),
                SelectionTuple(controller='controller_2', selection='catalog_1'),
            ]
        )
        self.central_controller.set_configuration(config_to_set)
        self.assertEqual(self.central_controller.get_configuration(), config_to_set)

        # Test for incorrect configuration
        with self.assertRaises(BiogemeError):
            self.central_controller.set_configuration(
                Configuration([SelectionTuple(controller='c4', selection='wrong')])
            )

    def test_prepare_operators(self):
        # Add test cases for prepare_operators
        operators = self.test_central_controller.prepare_operators()

        # Make sure the expected number of operators is generated
        #    Increase: 3
        #    Decrease: 3
        #    Pairs: 3 * 2 * 4 = 24
        #    Several increase: 1
        #    Several increase: 1
        #    Total: 32
        self.assertEqual(len(operators), 32)
        # Test a specific operator function, e.g., 'Increase c1'
        operator: ControllerOperator = operators['Increase c1']
        current_config = Configuration.from_string('c1:option1;c2:optionA;c3:choiceX')
        new_config, steps = operator(current_config=current_config, step=1)
        self.assertEqual(str(new_config), 'c1:option2;c2:optionA;c3:choiceX')
        self.assertEqual(steps, 1)

    def test_two_controllers(self):
        # Add test cases for two_controllers
        # Test for NE direction
        current_config = Configuration.from_string('c1:option1;c2:optionA;c3:choiceX')
        new_config, steps = self.test_central_controller.two_controllers(
            first_controller_name='c1',
            second_controller_name='c2',
            direction='NE',
            current_config=current_config,
            step=1,
        )
        self.assertEqual(str(new_config), 'c1:option2;c2:optionB;c3:choiceX')
        self.assertEqual(steps, 1)

        # Test for invalid direction
        with self.assertRaises(BiogemeError):
            self.central_controller.two_controllers(
                first_controller_name='c1',
                second_controller_name='c2',
                direction='invalid',
                current_config=current_config,
                step=1,
            )

    def test_modify_random_controllers(self):
        # Add test cases for modify_random_controllers
        # Test increasing controllers
        current_config = Configuration.from_string('c1:option1;c2:optionA;c3:choiceX')
        new_config, steps = self.test_central_controller.modify_random_controllers(
            current_config=current_config, increase=True, step=2
        )
        self.assertEqual(steps, 2)

        # Test decreasing controllers
        new_config, steps = self.test_central_controller.modify_random_controllers(
            current_config=current_config, increase=False, step=2
        )
        self.assertEqual(steps, 2)

    def test_set_controller(self):
        # Add test cases for set_controller
        # Test setting c1 to index 1
        self.test_central_controller.set_controller('c1', 1)
        expected_config = Configuration(
            [
                SelectionTuple(controller='c1', selection='option2'),
                SelectionTuple(controller='c2', selection='optionA'),
                SelectionTuple(controller='c3', selection='choiceX'),
            ]
        )
        self.assertEqual(
            self.test_central_controller.get_configuration(), expected_config
        )

        # Test setting an unknown controller
        with self.assertRaises(BiogemeError):
            self.test_central_controller.set_controller('c4', 1)

    def test_decreased_controller(self):
        # Add test cases for decreased_controller
        # Test decreasing c1 by 1
        current_config = Configuration.from_string('c1:option1;c2:optionA;c3:choiceX')
        new_config, steps = self.test_central_controller.decreased_controller(
            current_config=current_config, controller_name='c1', step=1
        )
        self.assertEqual(str(new_config), 'c1:option2;c2:optionA;c3:choiceX')
        self.assertEqual(steps, 1)

        # Test decreasing an unknown controller
        with self.assertRaises(BiogemeError):
            self.test_central_controller.decreased_controller(
                controller_name='c4', current_config=current_config, step=1
            )

    def test_expression_iterator(self):
        for named_expression in self.test_central_controller.expression_iterator():
            self.assertIn(
                named_expression.name,
                self.test_central_controller.all_configurations_ids,
            )

        for named_expression in self.central_controller.expression_iterator():
            self.assertIn(
                named_expression.name,
                self.central_controller.all_configurations_ids,
            )



"""Class that provides some automatic specification for segmented parameters

Michel Bierlaire
Thu Apr 3 12:07:44 2025
"""

from __future__ import annotations
from collections.abc import Iterable
from itertools import product
from typing import NamedTuple, TYPE_CHECKING

import pandas as pd

from .one_segmentation import OneSegmentation
from .segmentation_context import DiscreteSegmentationTuple
from biogeme.exceptions import BiogemeError
from biogeme.results_processing import EstimationResults

if TYPE_CHECKING:
    from biogeme.expressions import Beta, Expression, MultipleSum


class Segmentation:
    """Segmentation of a parameter, possibly with multiple socio-economic variables"""

    def __init__(
        self,
        beta: Beta,
        segmentation_tuples: Iterable[DiscreteSegmentationTuple],
        prefix: str = 'segmented',
    ):
        """Ctor

        :param beta: parameter to be segmented
        :param segmentation_tuples: characterization of the segmentations
        :param prefix: prefix to be used to generate the name of the
            segmented parameter
        """
        segmentation_tuples = tuple(segmentation_tuples)
        if not segmentation_tuples:
            raise BiogemeError('segmentation_tuples cannot be empty')
        self.beta: Beta = beta
        self.segmentations: tuple[OneSegmentation, ...] = tuple(
            OneSegmentation(beta, s) for s in segmentation_tuples
        )
        self.prefix = prefix

    def get_beta_ref_name(self) -> str:
        """
        Add a suffix to the name of the parameter
        """
        return self.beta.name + '_ref'

    def beta_ref_code(self) -> str:
        """Constructs the Python code for the parameter

        :return: Python code
        :rtype: str
        """

        beta_name = f"'{self.get_beta_ref_name()}'"
        return (
            f'Beta({beta_name}, {self.beta.init_value}, {self.beta.lower_bound}, '
            f'{self.beta.upper_bound}, {self.beta.status})'
        )

    def get_reference_beta(self) -> Beta:
        """Obtain the reference beta"""
        from biogeme.expressions import Beta

        beta_name = self.get_beta_ref_name()
        return Beta(
            beta_name,
            self.beta.init_value,
            self.beta.lower_bound,
            self.beta.upper_bound,
            self.beta.status,
        )

    def segmented_beta(self) -> Expression:
        """Create an expressions that combines all the segments

        :return: combined expression
        :rtype: biogeme.expressions.Expression

        """
        from biogeme.expressions import MultipleSum

        ref_beta = self.get_reference_beta()
        terms = [ref_beta]
        terms += [
            element for s in self.segmentations for element in s.list_of_expressions()
        ]

        return MultipleSum(terms)

    def segmented_code(self) -> str:
        """Create the Python code for an expressions that combines all the segments

        :return: Python code for the combined expression
        :rtype: str
        """
        result = '\n'.join(
            [
                s.beta_code(c, assignment=True)
                for s in self.segmentations
                for c in s.mapping.values()
            ]
        )
        result += '\n'

        terms = [self.beta_ref_code()]
        terms += [element for s in self.segmentations for element in s.list_of_code()]

        if len(terms) == 1:
            result += terms[0]
        else:
            joined_terms = ', '.join(terms)
            result += f'{self.prefix}_{self.beta.name} = bioMultSum([{joined_terms}])'
        return result

    def calculates_estimated_values(
        self, estimation_results: EstimationResults
    ) -> pd.DataFrame:
        """Calculates the estimated values of the parameter for each segment.

        :param estimation_results: results of the estimation
        :return: a pandas data frame with the definition of the segments and the corresponding values for the
        coefficient
        """

        class SegmentationValue(NamedTuple):
            segmentation: OneSegmentation
            value: str

        all_segmentations = [
            list(
                SegmentationValue(segment, value)
                for value in segment.segmentation_tuple.mapping.values()
            )
            for segment in self.segmentations
        ]

        # Use itertools.product to generate all combinations
        beta_values = estimation_results.get_beta_values()
        ref_beta_name = self.get_reference_beta().name
        ref_value = beta_values[ref_beta_name]
        list_of_rows = []
        for combination in product(*all_segmentations):
            the_row = {
                element.segmentation.variable.name: element.value
                for element in combination
            }
            the_row['parameter estimate'] = ref_value
            for element in combination:
                the_name = element.segmentation.beta_name(category=element.value)
                if the_name != ref_beta_name:
                    the_value = beta_values[the_name]
                    the_row['parameter estimate'] += the_value
            list_of_rows.append(the_row)
        df = pd.DataFrame(list_of_rows)
        return df



import unittest

import biogeme.biogeme as bio
import biogeme.distributions as dist
from biogeme import models
from biogeme.data.swissmetro import (
    CAR_AV_SP,
    CAR_CO_SCALED,
    CAR_TT_SCALED,
    CHOICE,
    GA,
    PURPOSE,
    SM_AV,
    SM_CO,
    SM_COST_SCALED,
    SM_TT_SCALED,
    TRAIN_AV_SP,
    TRAIN_CO,
    TRAIN_COST_SCALED,
    TRAIN_TT_SCALED,
    read_data,
)
from biogeme.expressions import Beta, IntegrateNormal, RandomVariable, exp, log

database = read_data()
# Keep only trip purposes 1 (commuter) and 3 (business)
exclude = ((PURPOSE != 1) * (PURPOSE != 3)) > 0
database.remove(exclude)


ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_TIME_S = Beta('B_TIME_S', 1, 0.0001, 10, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# Define a random parameter, normally distirbuted, designed to be used
# for Monte-Carlo simulation

omega = RandomVariable('omega')
B_TIME_RND = -exp(B_TIME + B_TIME_S * omega)
density = dist.normalpdf(omega)


# Utility functions

# If the person has a GA (season ticket) her incremental cost is actually 0
# rather than the cost value gathered from the
# network data.
SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)


V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}


av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# The choice model is a logit, with availability conditions
condprob = models.logit(V, av, CHOICE)
prob = IntegrateNormal(condprob, 'omega')
logprob = log(prob)


class test_17(unittest.TestCase):
    def testEstimation(self):
        biogeme = bio.BIOGEME(
            database,
            logprob,
            parameters=None,
            save_iterations=False,
            generate_html=False,
            generate_yaml=False,
        )
        results = biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, -5231.50634765625, 2)


if __name__ == '__main__':
    unittest.main()



import unittest

import biogeme.biogeme as bio
from biogeme import models
from biogeme.data.swissmetro import (
    CAR_AV_SP,
    CAR_CO_SCALED,
    CAR_TT_SCALED,
    CHOICE,
    GA,
    PURPOSE,
    SM_AV,
    SM_CO,
    SM_COST_SCALED,
    SM_TT_SCALED,
    TRAIN_AV_SP,
    TRAIN_CO,
    TRAIN_COST_SCALED,
    TRAIN_TT_SCALED,
    read_data,
)
from biogeme.expressions import Beta
from biogeme.nests import NestsForNestedLogit, OneNestForNestedLogit

database = read_data()
# Keep only trip purposes 1 (commuter) and 3 (business)
exclude = ((PURPOSE != 1) * (PURPOSE != 3)) > 0
database.remove(exclude)


ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# Note that MU cannot be zero
MU = Beta('MU', 0.5, 0.000001, 1.0, 0)


SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)


V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}


av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# Definition of nests:
# 1: nests parameter
# 2: list of alternatives

existing = OneNestForNestedLogit(
    nest_param=1, list_of_alternatives=[1, 3], name='existing'
)
nests = NestsForNestedLogit(choice_set=[1, 2, 3], tuple_of_nests=(existing,))

# The choice model is a nested logit, with availability conditions
log_prob = models.lognested_mev_mu(V, av, nests, CHOICE, MU)


class test_10(unittest.TestCase):
    def testEstimation(self):
        biogeme = bio.BIOGEME(
            database,
            log_prob,
            parameters=None,
            save_iterations=False,
            generate_html=False,
            generate_yaml=False,
        )
        results = biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, -5236.90001357885, 2)


if __name__ == '__main__':
    unittest.main()



"""Implementation of the "non-monotonic" MDCEV model. See the technical report.

Michel Bierlaire
Tue Apr 9 09:25:20 2024
"""

from functools import lru_cache

import numpy as np

from biogeme.calculator import evaluate_expression
from biogeme.database import Database
from biogeme.exceptions import BiogemeError
from biogeme.expressions import Expression, exp, log
from biogeme.mdcev.mdcev import Mdcev
from biogeme.tools.checks import validate_dict_types


class NonMonotonic(Mdcev):
    """Implementation of the MDCEV model with non-monotonic utility function."""

    def __init__(
        self,
        model_name: str,
        baseline_utilities: dict[int, Expression],
        gamma_parameters: dict[int, Expression | None],
        mu_utilities: dict[int, Expression],
        alpha_parameters: dict[int, Expression] | None = None,
        scale_parameter: Expression | None = None,
        weights: Expression | None = None,
    ) -> None:
        """Implementation of the MDCEV model with Gamma profile."""
        super().__init__(
            model_name=model_name,
            baseline_utilities=baseline_utilities,
            gamma_parameters=gamma_parameters,
            alpha_parameters=alpha_parameters,
            scale_parameter=scale_parameter,
            weights=weights,
        )
        self.mu_utilities: dict[int, Expression] = mu_utilities
        validate_dict_types(self.mu_utilities, 'mu_utilities', Expression)

    def transformed_utility(
        self,
        the_id: int,
        the_consumption: Expression,
    ) -> Expression:
        """Calculates the utility for one alternative. ."""

        if the_id not in self.alternatives:
            error_msg = (
                f'Alternative id {the_id} is invalid. Valid ids: {self.alternatives}'
            )
            raise BiogemeError(error_msg)
        gamma = self.gamma_parameters[the_id]

        baseline_utility = self.baseline_utilities[the_id]
        mu_utility = self.mu_utilities[the_id]

        alpha_parameter = self.alpha_parameters[the_id]

        if gamma is None:
            return mu_utility + exp(baseline_utility) * the_consumption ** (
                alpha_parameter - 1
            )

        return mu_utility + exp(baseline_utility) * (1 + the_consumption / gamma) ** (
            alpha_parameter - 1
        )

    def calculate_log_determinant_one_alternative(
        self, the_id: int, consumption: Expression
    ) -> Expression:
        """Calculate the log of the entries for the determinant. For
            the outside good, gamma is equal to 0.

        :param the_id: identifier of the good.
        :param consumption: expression for the consumption.
        """
        if the_id not in self.alternatives:
            error_msg = (
                f'Alternative id {the_id} is invalid. Valid ids: {self.alternatives}'
            )
            raise BiogemeError(error_msg)
        gamma = self.gamma_parameters[the_id]

        if gamma is None:
            return (
                self.baseline_utilities[the_id]
                + log(1 - self.alpha_parameters[the_id])
                + (self.alpha_parameters[the_id] - 2) * log(consumption)
            )

        return (
            self.baseline_utilities[the_id]
            + log(1 - self.alpha_parameters[the_id])
            - log(gamma)
            + (self.alpha_parameters[the_id] - 2) * log(1 + consumption / gamma)
        )

    def calculate_inverse_of_determinant_one_alternative(
        self, the_id: int, consumption: Expression
    ) -> Expression:
        """Calculate the inverse of the entries for the determinant. For
            the outside good, gamma is equal to 0.

        :param the_id: identifier of the good.
        :param consumption: expression for the consumption.
        """
        if the_id not in self.alternatives:
            error_msg = (
                f'Alternative id {the_id} is invalid. Valid ids: {self.alternatives}'
            )
            raise BiogemeError(error_msg)
        gamma: Expression = self.gamma_parameters[the_id]

        if gamma is None:
            return (
                exp(-self.baseline_utilities[the_id])
                * consumption ** (2 - self.alpha_parameters[the_id])
                / (1 - self.alpha_parameters[the_id])
            )
        return (
            exp(-self.baseline_utilities[the_id])
            * gamma
            * (1 + consumption / gamma) ** (2 - self.alpha_parameters[the_id])
            / (1 - self.alpha_parameters[the_id])
        )

    @lru_cache
    def calculate_mu_utility(
        self, alternative_id: int, one_observation: Database
    ) -> float:
        """As this function may be called many times with the same input in forecasting mode, we use the
        lru_cache decorator."""
        assert one_observation.num_rows() == 1
        if self.estimation_results:
            return evaluate_expression(
                expression=self.mu_utilities[alternative_id],
                numerically_safe=False,
                database=one_observation,
                betas=self.estimation_results.get_beta_values(),
                aggregation=True,
            )

        return evaluate_expression(
            expression=self.mu_utilities[alternative_id],
            numerically_safe=False,
            database=one_observation,
            aggregation=True,
        )

    def utility_expression_one_alternative(
        self,
        the_id: int,
        the_consumption: Expression,
        unscaled_epsilon: Expression,
    ) -> Expression:
        """Utility expression. Used only for code validation."""
        baseline_utility = self.baseline_utilities[the_id]
        mu_utility = self.mu_utilities[the_id]
        epsilon: Expression = (
            unscaled_epsilon
            if self.scale_parameter is None
            else unscaled_epsilon / self.scale_parameter
        )
        alpha: Expression = self.alpha_parameters[the_id]
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            return (
                exp(baseline_utility) * (the_consumption**alpha) / alpha
                + (mu_utility + epsilon) * the_consumption
            )
        return (
            gamma
            * exp(baseline_utility)
            * ((1 + the_consumption / gamma) ** alpha - 1)
            / alpha
            + (mu_utility + epsilon) * the_consumption
        )

    def utility_one_alternative(
        self,
        the_id: int,
        the_consumption: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Utility needed for forecasting"""
        baseline_utility = self.calculate_baseline_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        mu_utility = self.calculate_mu_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        if self.scale_parameter is not None:
            epsilon /= self.scale_parameter.get_value()
        alpha: float = self.alpha_parameters[the_id].get_value()
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            return (
                np.exp(baseline_utility) * (the_consumption**alpha) / alpha
                + (mu_utility + epsilon) * the_consumption
            )
        return (
            gamma.get_value()
            * np.exp(baseline_utility)
            * ((1 + the_consumption / gamma.get_value()) ** alpha - 1)
            / alpha
            + (mu_utility + epsilon) * the_consumption
        )

    def derivative_utility_one_alternative(
        self,
        the_id: int,
        the_consumption: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Used in the optimization problem solved for forecasting tp calculate the dual variable."""
        baseline_utility = self.calculate_baseline_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        mu_utility = self.calculate_mu_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        if self.scale_parameter is not None:
            epsilon /= self.scale_parameter.get_value()
        alpha: float = self.alpha_parameters[the_id].get_value()
        gamma: Expression | None = self.gamma_parameters[the_id]
        if gamma is None:
            return (
                np.exp(baseline_utility) * (the_consumption ** (alpha - 1.0))
                + mu_utility
                + epsilon
            )
        return (
            np.exp(baseline_utility)
            * (1.0 + the_consumption / gamma.get_value()) ** (alpha - 1.0)
            + mu_utility
            + epsilon
        )

    @lru_cache
    def optimal_consumption_one_alternative(
        self,
        the_id: int,
        dual_variable: float,
        epsilon: float,
        one_observation: Database,
    ) -> float:
        """Analytical calculation of the optimal consumption if the dual variable is known."""
        baseline_utility = self.calculate_baseline_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        mu_utility = self.calculate_mu_utility(
            alternative_id=the_id, one_observation=one_observation
        )
        if self.scale_parameter is not None:
            epsilon /= self.scale_parameter.get_value()
        alpha: float = self.alpha_parameters[the_id].get_value()
        gamma: Expression | None = self.gamma_parameters[the_id]
        base = (dual_variable - mu_utility - epsilon) * np.exp(-baseline_utility)

        exponent = 1.0 / (alpha - 1.0)
        if gamma is None:
            return base**exponent
        result = gamma.get_value() * (base**exponent - 1)
        return result

    def _list_of_expressions(self) -> list[Expression]:
        """Extract the list of expressions involved in the model"""
        return [expression for expression in self.mu_utilities.values()]

    def lower_bound_dual_variable(
        self,
        chosen_alternatives: set[int],
        one_observation: Database,
        epsilon: np.ndarray,
    ) -> float:
        """Method providing model specific lower bound on the dual variable.

        :param chosen_alternatives: list of alternatives that are chosen at the optimal solution
        :param one_observation: data for one observation.
        :param epsilon: draws from the error term.
        :return: a lower bound and upper bound on the dual variable
        """
        lower_bound = -np.inf
        for alternative_id in chosen_alternatives:
            epsilon_alternative = epsilon[self.key_to_index[alternative_id]]
            if self.scale_parameter is not None:
                epsilon_alternative /= self.scale_parameter.get_value()
            mu_utility = self.calculate_mu_utility(
                alternative_id=alternative_id, one_observation=one_observation
            )
            mu_utility += epsilon_alternative
            if mu_utility > lower_bound:
                lower_bound = mu_utility
        return lower_bound



"""

List of alternatives
====================

Script reading the list of alternatives and identifying subsets

:author: Michel Bierlaire
:date: Mon Oct  9 10:53:03 2023
"""

import pandas as pd
from biogeme.partition import Partition

# %%
alternatives = pd.read_csv('restaurants.dat')
alternatives

# %%
ID_COLUMN = 'ID'

# %%
all_alternatives = set(list(alternatives[ID_COLUMN]))

# %%
# Set of Asian restaurants
asian = set(alternatives[alternatives['Asian'] == 1][ID_COLUMN])
print(f'Number of asian restaurants: {len(asian)}')

# %%
# Set of restaurants located in downtown
downtown = set(alternatives[alternatives['downtown'] == 1][ID_COLUMN])

# %%
# Set of Asian restaurants in downtown
asian_and_downtown = asian & downtown

# %%
# Set of Asian restaurants, and of restaurants in downtown
asian_or_downtown = asian | downtown

# %%
# Set of Asian restaurants not in downtown
only_asian = asian - asian_and_downtown

# %%
# Set of non Asian restaurants in downtown
only_downtown = downtown - asian_and_downtown

# %%
# Set of restaurants that are neither Asian nor in downtown
others = all_alternatives - asian_or_downtown


# %%
def complement(a_set: set[int]) -> set[int]:
    """Returns the complement of a set"""
    return all_alternatives - a_set


# %%
# Partitions.
partition_asian = Partition([asian, complement(asian)], full_set=all_alternatives)
partition_downtown = Partition(
    [downtown, complement(downtown)], full_set=all_alternatives
)
partition_uniform = Partition([all_alternatives], full_set=all_alternatives)
partition_uniform_asian = Partition([asian], full_set=asian)
partition_uniform_asian_or_downtown = Partition(
    [asian_or_downtown], full_set=asian_or_downtown
)

# %%
partitions = {
    'uniform': partition_uniform,
    'asian': partition_asian,
    'downtown': partition_downtown,
    'uniform_asian_or_downtown': partition_uniform_asian_or_downtown,
    'uniform_asian': partition_uniform_asian,
}



"""
Identification of Pareto optimal models

Michel Bierlaire
Fri Oct 4 09:22:22 2024
"""

from biogeme_optimization import pareto
from biogeme_optimization.pareto import Pareto
from .estimation_results import EstimationResults


def pareto_optimal(
    dict_of_results: dict[str, EstimationResults], a_pareto: Pareto | None = None
) -> dict[str, EstimationResults]:
    """Identifies the non dominated models, with respect to maximum
    log likelihood and minimum number of parameters

    :param dict_of_results: dict of results associated with their config ID
    :param a_pareto: if not None, Pareto set where the results will be inserted.
    :return: a dict of named results with pareto optimal results
    """
    if a_pareto is None:
        the_pareto = pareto.Pareto()
    else:
        the_pareto = a_pareto
    for config_id, estimation_results in dict_of_results.items():
        the_element = pareto.SetElement(
            element_id=config_id,
            objectives=[
                -estimation_results.final_log_likelihood,
                estimation_results.number_of_parameters,
            ],
        )
        the_pareto.add(the_element)

    selected_results = {
        element.element_id: dict_of_results[element.element_id]
        for element in the_pareto.pareto
    }
    the_pareto.dump()
    return selected_results



"""Arithmetic expressions accepted by Biogeme: draws

Michel Bierlaire
Fri Jun 27 2025, 14:41:17
"""

from __future__ import annotations

import logging

import jax.numpy as jnp
from biogeme.exceptions import BiogemeError

from .elementary_expressions import Elementary
from .elementary_types import TypeOfElementaryExpression
from .jax_utils import JaxFunctionType

logger = logging.getLogger(__name__)


class Draws(Elementary):
    """
    Draws for Monte-Carlo integration
    """

    expression_type = TypeOfElementaryExpression.DRAWS

    def __init__(self, name: str, draw_type: str):
        """Constructor

        :param name: name of the random variable with a series of draws.
        :type name: string
        :param draw_type: type of draws.
        :type draw_type: string
        """
        super().__init__(name)
        self.draw_type = draw_type

    def deep_flat_copy(self) -> Draws:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        return type(self)(name=self.name, draw_type=self.draw_type)

    def __str__(self) -> str:
        return f'Draws("{self.name}", "{self.draw_type}")'

    @property
    def safe_draw_id(self) -> int:
        """Check the presence of the draw ID before its usage"""
        if self.specific_id is None:
            raise BiogemeError(f"No id defined for draw {self.name}")
        return self.specific_id

    def dict_of_elementary_expression(
        self, the_type: TypeOfElementaryExpression
    ) -> dict[str, Elementary]:
        """Extract a dict with all elementary expressions of a specific type

        :param the_type: the type of expression
        :type  the_type: TypeOfElementaryExpression
        """

        if the_type == TypeOfElementaryExpression.DRAWS:
            # Until version 3.2.13, this function returned the following:
            # return {self.name: self.drawType}
            return {self.name: self}
        return {}

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.ndarray:
            return jnp.take(the_draws, self.safe_draw_id, axis=-1)

        return the_jax_function



"""
.. _optima_data:

Data preparation for Optima
===========================

Prepare data for the Optima case study.

:author: Michel Bierlaire
:date: Wed Apr 12 20:52:37 2023

"""

import os

import pandas as pd

import biogeme.database as db
from biogeme.expressions import Variable

# Get the directory of the current file
module_dir = os.path.dirname(__file__)
# Construct the path to the data file
data_file_path = os.path.join(module_dir, 'data', 'optima.dat')


# %%
# Read the data
def read_data() -> db.Database:
    """Read the data from file"""
    df = pd.read_csv(data_file_path, sep='\t')
    # Exclude observations such that the chosen alternative is -1
    df.drop(df[df['Choice'] == -1].index, inplace=True)

    car_not_available = df['CarAvail'] == 3
    car_is_chosen = df['Choice'] == 1
    incompatible = car_is_chosen & car_not_available
    df.drop(df[incompatible].index, inplace=True)

    # Normalize the weights
    sum_weight = df['Weight'].sum()
    number_of_rows = df.shape[0]
    df['normalized_weight'] = df['Weight'] * number_of_rows / sum_weight
    database = db.Database(name=data_file_path, dataframe=df)
    _ = database.define_variable('ScaledIncome', CalculatedIncome / 1000)
    _ = database.define_variable('age_65_more', age >= 65)
    _ = database.define_variable('moreThanOneCar', NbCar > 1)
    _ = database.define_variable('moreThanOneBike', NbBicy > 1)
    _ = database.define_variable('individualHouse', HouseType == 1)
    _ = database.define_variable('male', Gender == 1)
    _ = database.define_variable(
        'haveChildren', ((FamilSitu == 3) + (FamilSitu == 4)) > 0
    )
    _ = database.define_variable('haveGA', GenAbST == 1)
    _ = database.define_variable('highEducation', Education >= 6)
    _ = database.define_variable(
        'childCenter', ((ResidChild == 1) + (ResidChild == 2)) > 0
    )

    _ = database.define_variable(
        'childSuburb', ((ResidChild == 3) + (ResidChild == 4)) > 0
    )
    _ = database.define_variable('TimePT_scaled', TimePT / 200)
    _ = database.define_variable('TimeCar_scaled', TimeCar / 200)
    _ = database.define_variable('MarginalCostPT_scaled', MarginalCostPT / 10)
    _ = database.define_variable('CostCarCHF_scaled', CostCarCHF / 10)
    _ = database.define_variable('distance_km_scaled', distance_km / 5)
    _ = database.define_variable('PurpHWH', TripPurpose == 1)
    _ = database.define_variable('PurpOther', TripPurpose != 1)

    import numpy as np

    nan_positions = np.where(database.dataframe.isna())
    for row, col in zip(*nan_positions):
        ic(f"NaN at row {row}, column '{database.dataframe.columns[col]}'")
    return database


# %%
# Variables from the data
Choice = Variable('Choice')
TimePT = Variable('TimePT')
TimeCar = Variable('TimeCar')
MarginalCostPT = Variable('MarginalCostPT')
CostCarCHF = Variable('CostCarCHF')
distance_km = Variable('distance_km')
Gender = Variable('Gender')
OccupStat = Variable('OccupStat')
Weight = Variable('Weight')
ID = Variable('ID')
DestAct = Variable('DestAct')
NbTransf = Variable('NbTransf')
WalkingTimePT = Variable('WalkingTimePT')
WaitingTimePT = Variable('WaitingTimePT')
CostPT = Variable('CostPT')
CostCar = Variable('CostCar')
NbHousehold = Variable('NbHousehold')
NbChild = Variable('NbChild')
NbCar = Variable('NbCar')
NbMoto = Variable('NbMoto')
NbBicy = Variable('NbBicy')
NbBicyChild = Variable('NbBicyChild')
NbComp = Variable('NbComp')
NbTV = Variable('NbTV')
Internet = Variable('Internet')
NewsPaperSubs = Variable('NewsPaperSubs')
NbCellPhones = Variable('NbCellPhones')
NbSmartPhone = Variable('NbSmartPhone')
HouseType = Variable('HouseType')
OwnHouse = Variable('OwnHouse')
NbRoomsHouse = Variable('NbRoomsHouse')
YearsInHouse = Variable('YearsInHouse')
Income = Variable('Income')
BirthYear = Variable('BirthYear')
Mothertongue = Variable('Mothertongue')
FamilSitu = Variable('FamilSitu')
SocioProfCat = Variable('SocioProfCat')
CalculatedIncome = Variable('CalculatedIncome')
Education = Variable('Education')
HalfFareST = Variable('HalfFareST')
LineRelST = Variable('LineRelST')
GenAbST = Variable('GenAbST')
AreaRelST = Variable('AreaRelST')
OtherST = Variable('OtherST')
CarAvail = Variable('CarAvail')
Envir01 = Variable('Envir01')
Envir02 = Variable('Envir02')
Envir03 = Variable('Envir03')
Envir04 = Variable('Envir04')
Envir05 = Variable('Envir05')
Envir06 = Variable('Envir06')
Mobil01 = Variable('Mobil01')
Mobil02 = Variable('Mobil02')
Mobil03 = Variable('Mobil03')
Mobil04 = Variable('Mobil04')
Mobil05 = Variable('Mobil05')
Mobil06 = Variable('Mobil06')
Mobil07 = Variable('Mobil07')
Mobil08 = Variable('Mobil08')
Mobil09 = Variable('Mobil09')
Mobil10 = Variable('Mobil10')
Mobil11 = Variable('Mobil11')
Mobil12 = Variable('Mobil12')
Mobil13 = Variable('Mobil13')
Mobil14 = Variable('Mobil14')
Mobil15 = Variable('Mobil15')
Mobil16 = Variable('Mobil16')
Mobil17 = Variable('Mobil17')
Mobil18 = Variable('Mobil18')
Mobil19 = Variable('Mobil19')
Mobil20 = Variable('Mobil20')
Mobil21 = Variable('Mobil21')
Mobil22 = Variable('Mobil22')
Mobil23 = Variable('Mobil23')
Mobil24 = Variable('Mobil24')
Mobil25 = Variable('Mobil25')
Mobil26 = Variable('Mobil26')
Mobil27 = Variable('Mobil27')
ResidCh01 = Variable('ResidCh01')
ResidCh02 = Variable('ResidCh02')
ResidCh03 = Variable('ResidCh03')
ResidCh04 = Variable('ResidCh04')
ResidCh05 = Variable('ResidCh05')
ResidCh06 = Variable('ResidCh06')
ResidCh07 = Variable('ResidCh07')
LifSty01 = Variable('LifSty01')
LifSty02 = Variable('LifSty02')
LifSty03 = Variable('LifSty03')
LifSty04 = Variable('LifSty04')
LifSty05 = Variable('LifSty05')
LifSty06 = Variable('LifSty06')
LifSty07 = Variable('LifSty07')
LifSty08 = Variable('LifSty08')
LifSty09 = Variable('LifSty09')
LifSty10 = Variable('LifSty10')
LifSty11 = Variable('LifSty11')
LifSty12 = Variable('LifSty12')
LifSty13 = Variable('LifSty13')
LifSty14 = Variable('LifSty14')
TripPurpose = Variable('TripPurpose')
TypeCommune = Variable('TypeCommune')
UrbRur = Variable('UrbRur')
LangCode = Variable('LangCode')
ClassifCodeLine = Variable('ClassifCodeLine')
frequency = Variable('frequency')
ResidChild = Variable('ResidChild')
NbTrajects = Variable('NbTrajects')
FreqCarPar = Variable('FreqCarPar')
FreqTrainPar = Variable('FreqTrainPar')
FreqOtherPar = Variable('FreqOtherPar')
FreqTripHouseh = Variable('FreqTripHouseh')
Region = Variable('Region')
InVehicleTime = Variable('InVehicleTime')
ModeToSchool = Variable('ModeToSchool')
ReportedDuration = Variable('ReportedDuration')
CoderegionCAR = Variable('CoderegionCAR')
age = Variable('age')
normalized_weight = Variable('normalized_weight')

ScaledIncome = Variable('ScaledIncome')
age_65_more = Variable('age_65_more')
moreThanOneCar = Variable('moreThanOneCar')
moreThanOneBike = Variable('moreThanOneBike')
individualHouse = Variable('individualHouse')
male = Variable('male')
haveChildren = Variable('haveChildren')
haveGA = Variable('haveGA')
highEducation = Variable('highEducation')
childCenter = Variable('childCenter')
childSuburb = Variable('childSuburb')
TimePT_scaled = Variable('TimePT_scaled')
TimeCar_scaled = Variable('TimeCar_scaled')
MarginalCostPT_scaled = Variable('MarginalCostPT_scaled')
CostCarCHF_scaled = Variable('CostCarCHF_scaled')
distance_km_scaled = Variable('distance_km_scaled')
PurpHWH = Variable('PurpHWH')
PurpOther = Variable('PurpOther')



import unittest

import numpy as np

import biogeme.biogeme as bio
from biogeme import models
from biogeme.data.swissmetro import (
    CAR_AV_SP,
    CAR_CO_SCALED,
    CAR_TT_SCALED,
    CHOICE,
    GA,
    PURPOSE,
    SM_AV,
    SM_CO,
    SM_COST_SCALED,
    SM_TT_SCALED,
    TRAIN_AV_SP,
    TRAIN_CO,
    TRAIN_COST_SCALED,
    TRAIN_TT_SCALED,
    read_data,
)
from biogeme.draws import RandomNumberGeneratorTuple
from biogeme.expressions import Beta, Draws, MonteCarlo, log
from biogeme.parameters import Parameters

database = read_data()
# Keep only trip purposes 1 (commuter) and 3 (business)
exclude = ((PURPOSE != 1) * (PURPOSE != 3)) > 0
database.remove(exclude)


ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_TIME_S = Beta('B_TIME_S', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# Define a random parameter with a triangular distribution, designed to be used
# for Monte-Carlo simulation


# Provide my own random number generator to the database.
# See the numpy.random documentation to obtain a list of other distributions.
def the_triangular_generator(sample_size, number_of_draws):
    return np.random.triangular(-1, 0, 1, (sample_size, number_of_draws))


my_random_number_generators = {
    'TRIANGULAR': RandomNumberGeneratorTuple(
        generator=the_triangular_generator,
        description='Triangular distribution T(-1,0,1)',
    )
}

B_TIME_RND = B_TIME + B_TIME_S * Draws('b_time_rnd', 'TRIANGULAR')

# Utility functions

# If the person has a GA (season ticket) her incremental cost is actually 0
# rather than the cost value gathered from the
# network data.
SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)


V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}


av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# The choice model is a logit, with availability conditions
prob = models.logit(V, av, CHOICE)
logprob = log(MonteCarlo(prob))


class test_25(unittest.TestCase):
    def testEstimation(self):
        parameters = Parameters()
        parameters.set_value(section='MonteCarlo', name='number_of_draws', value=5)
        parameters.set_value(section='MonteCarlo', name='seed', value=10)
        biogeme = bio.BIOGEME(
            database,
            logprob,
            parameters=parameters,
            random_number_generators=my_random_number_generators,
            save_iterations=False,
            generate_html=False,
            generate_yaml=False,
        )
        results = biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, -5309.388413432725, 2)


if __name__ == '__main__':
    unittest.main()



"""

Specification of the mixtures of logit
======================================

Creation of the Biogeme object for a mixtures of logit models where
the integral is approximated using MonteCarlo integration.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 21:04:47 2023
"""

import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, MonteCarlo, log, bioDraws
from biogeme.tools import TemporaryFile

from swissmetro import (
    database,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    TRAIN_AV_SP,
    SM_AV,
    CAR_AV_SP,
    CHOICE,
)

# %%
R = 2000


# %%
def get_biogeme(the_draws: bioDraws, number_of_draws: int) -> bio.BIOGEME:
    """Function returning the Biogeme object as a function of the selected draws

    :param the_draws: expression representing the draws.
    :param number_of_draws: number of draws to generate.
    :return: Biogeme object.
    """

    asc_car = Beta('asc_car', 0, None, None, 0)
    asc_train = Beta('asc_train', 0, None, None, 0)
    b_time = Beta('b_time', 0, None, None, 0)
    b_time_s = Beta('b_time_s', 1, None, None, 0)
    b_cost = Beta('b_cost', 0, None, None, 0)

    # Define a random parameter, normally distributed, designed to be used
    # for Monte-Carlo simulation
    b_time_rnd = b_time + b_time_s * the_draws

    # Definition of the utility functions
    v_1 = asc_train + b_time_rnd * TRAIN_TT_SCALED + b_cost * TRAIN_COST_SCALED
    v_2 = b_time_rnd * SM_TT_SCALED + b_cost * SM_COST_SCALED
    v_3 = asc_car + b_time_rnd * CAR_TT_SCALED + b_cost * CAR_CO_SCALED

    # Associate utility functions with the numbering of alternatives
    utilities = {1: v_1, 2: v_2, 3: v_3}

    # Associate the availability conditions with the alternatives
    av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

    # The choice model is a logit, with availability conditions
    prob = models.logit(utilities, av, CHOICE)
    logprob = log(MonteCarlo(prob))

    with TemporaryFile() as filename:
        with open(filename, 'w', encoding='utf-8') as f:
            print('[MonteCarlo]', file=f)
            print(f'number_of_draws = {number_of_draws}', file=f)

        the_biogeme = bio.BIOGEME(database, logprob, parameter_file=filename)

    return the_biogeme



""" Functions to calculate the log likelihood

:author: Michel Bierlaire
:date: Fri Mar 29 17:11:44 2019

"""

# Too constraining
# pylint: disable=invalid-name,

from biogeme.expressions import Expression, exp, log, MonteCarlo


def loglikelihood(prob: Expression) -> Expression:
    """
    Simply computes the log of the probability

    :param prob: An expression providing the value of the probability.

    :return: the logarithm of the probability.

    """
    return log(prob)


def mixedloglikelihood(prob: Expression) -> Expression:
    """Compute a simulated loglikelihood function

    :param prob: An expression providing the value of the
                 probability. Although it is not formally necessary,
                 the expression should contain one or more random
                 variables of a given distribution, and therefore
                 is defined as

    .. math:: P(i|\\xi_1,\\ldots,\\xi_L)


    :return: the simulated loglikelihood, given by

        .. math:: \\ln\\left(\\sum_{r=1}^R
            P(i|\\xi^r_1,\\ldots,\\xi^r_L) \\right)

        where :math:`R` is the number of draws, and :math:`\\xi_j^r`
        is the rth draw of the random variable :math:`\\xi_j`.

    """
    ell = MonteCarlo(prob)
    return log(ell)


def likelihoodregression(
    meas: Expression, model: Expression, sigma: Expression
) -> Expression:
    """Computes likelihood function of a regression model.

    :param meas: An expression providing the value :math:`y` of the measure
                 for the current observation.

    :param model: An expression providing the output :math:`m` of the model
                  for the current observation.

    :param sigma: An expression (typically, a parameter) providing the
                  standard error :math:`\\sigma` of the error term.
    :type sigma: biogeme.expressions.Expression
    :return: The likelihood of the regression, assuming a normal distribution,
        that is

        .. math:: \\frac{1}{\\sigma} \\phi\\left( \\frac{y-m}{\\sigma} \\right)

        where :math:`\\phi(\\cdot)` is the pdf of the normal distribution.

    """
    return exp(loglikelihoodregression(meas, model, sigma))


def loglikelihoodregression(
    meas: Expression, model: Expression, sigma: Expression
) -> Expression:
    """Computes log likelihood function of a regression model.

    :param meas: An expression providing the value :math:`y` of the
                 measure for the current observation.

    :param model: An expression providing the output :math:`m` of the
                  model for the current observation.

    :param sigma: An expression (typically, a parameter) providing
                  the standard error :math:`\\sigma` of the error term.

    :return: the likelihood of the regression, assuming a normal distribution,
        that is

    .. math:: -\\left( \\frac{(y-m)^2}{2\\sigma^2} \\right) -
              \\frac{1}{2}\\log(\\sigma^2) - \\frac{1}{2}\\log(2\\pi)

    """
    t = (meas - model) / sigma
    f = -(t**2) / 2 - log(sigma**2) / 2 - 0.9189385332
    return f



"""Arithmetic expressions accepted by Biogeme: logzero

Michel Bierlaire
Sat Jun 28 2025, 12:17:05
"""

from __future__ import annotations

import logging

import jax
import jax.numpy as jnp
import numpy as np
from biogeme.floating_point import EPSILON, JAX_FLOAT

from .base_expressions import ExpressionOrNumeric
from .jax_utils import JaxFunctionType
from .unary_expressions import UnaryOperator

logger = logging.getLogger(__name__)


class logzero(UnaryOperator):
    """
    logarithm expression. Returns zero if the argument is zero.
    """

    def __init__(self, child: ExpressionOrNumeric):
        """Constructor

        :param child: first arithmetic expression
        :type child: biogeme.expressions.Expression
        """
        super().__init__(child)

    def deep_flat_copy(self) -> logzero:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        copy_child = self.child.deep_flat_copy()
        return type(self)(child=copy_child)

    def __str__(self) -> str:
        return f'logzero({self.child})'

    def __repr__(self) -> str:
        return f'logzero({repr(self.child)})'

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        v = self.child.get_value()
        return 0 if v == 0 else np.log(v)

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a function to be used by biogeme_jax. Must be overloaded
            by each expression
        :return: the function takes two parameters: the parameters,
            and one row of the database.
        """
        child_jax = self.child.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )

        if numerically_safe:

            def the_jax_function(
                parameters: jnp.ndarray,
                one_row: jnp.ndarray,
                the_draws: jnp.ndarray,
                the_random_variables: jnp.ndarray,
            ) -> jnp.ndarray:
                child_value = child_jax(
                    parameters, one_row, the_draws, the_random_variables
                )
                is_zero = child_value == 0.0
                slope = 1.0 / EPSILON
                intercept = jnp.log(EPSILON) - slope * EPSILON
                approx_log = slope * child_value + intercept

                return jax.lax.cond(
                    is_zero,
                    lambda _: jnp.array(0.0, dtype=JAX_FLOAT),
                    lambda _: jax.lax.cond(
                        child_value < EPSILON,
                        lambda _: approx_log,
                        lambda _: jnp.log(child_value),
                        operand=None,
                    ),
                    operand=None,
                )

            return the_jax_function

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.ndarray:
            child_value = child_jax(
                parameters, one_row, the_draws, the_random_variables
            )
            is_zero = child_value == 0.0

            return jax.lax.cond(
                is_zero,
                lambda _: jnp.array(0.0, dtype=JAX_FLOAT),
                lambda _: jnp.log(child_value),
                operand=None,
            )

        return the_jax_function
